# Copyright (c) 2003-2019  Pavel Rychly, Vojtech Kovar, Milos Jakubicek, Milos Husak, Vit Baisa

from CGIPublisher import correct_types
from usercgi import UserCGI
import corplib, conclib, version
from corplib import corpconf_pairs
import os, re
import sys
import time
import glob
from butils import *

def onelevelcrit (prefix, attr, ctx, pos, fcode, icase, bward='', empty=''):
    fromcode = {'lc': '<0', 'rc': '>0', 'kl': '<0', 'kr': '>0'}
    attrpart = '%s%s/%s%s%s ' % (prefix, attr, icase, bward, empty)
    if not ctx:
        ctx = '%i%s' % (pos, fromcode.get (fcode, '0'))
    if '~' in ctx and '.' in attr:
        ctx = ctx.split('~')[0]
    return attrpart + ctx

def multilevel_sort_crit(mlsort_options):
    "multilevel sort for Platypus"
    prefix = 's'
    result = ''
    for opt in mlsort_options:
        if opt['ctx'] == '0': ctx = '0<0~0>0' # 0 == KWIC
        elif int(opt['ctx']) > 0: ctx = opt['ctx'] + '>0' # 0 == KWIC
        else: ctx = opt['ctx']
        result += onelevelcrit(prefix, opt['attr'], ctx, '', '',
                               opt.get('icase', ''), opt.get('bward', ''))
        prefix = ' '
    return result

def nicearg(arg):
    args = arg.split('"');
    niceargs = []
    niceargsset = set()
    for i in range(len(args)):
        if (i % 2):
            tmparg = args[i].strip('\\').replace('(?i)', '')
            if tmparg not in niceargsset:
                niceargs.append(tmparg)
                niceargsset.add(tmparg)
        else:
            if args[i].startswith('within'):
                niceargs.append('within')
    return ', '.join(niceargs)

def obfuscate_terms(l):
    import random
    # used in extract_{keywords,terms}
    j = 0
    x, y = 0, 0
    for i in range(len(l)):
        x = i / 5
        if y == x:
            y = 0
        else:
            obfs = ''
            for c in l[i]['item']:
                if c == ' ':
                    obfs += c
                else:
                    obfs += chr(random.randint(97,122))
            l[i]['item'] = obfs
            l[i]['obfs'] = '1'
            l[i]['frq1'] = 0
            l[i]['link'] = ''
            y += 1
    return l

def obfuscate_biterms(l, d):
    import random
    j, x, y = 0, 0, 0
    for i in range(len(l)):
        x = i / 5
        if y == x:
            y = 0
            for bdi in d[l[i]]:
                bdi.append(0)
        else:
            d[l[i]] = d[l[i]][:1] # no alternative translations
            obfs_l1 = ''
            obfs_l2 = ''
            bdi = d[l[i]][0]
            for c in bdi[0]:
                if c == ' ':
                    obfs_l1 += ' '
                else:
                    obfs_l1 += chr(random.randint(97,122))
            for c in bdi[1]:
                if c == ' ':
                    obfs_l2 += ' '
                else:
                    obfs_l2 += chr(random.randint(97,122))
            bdi[0] = obfs_l1
            bdi[1] = obfs_l2
            bdi[2] = bdi[3] = bdi[4] = 0
            d[obfs_l1] = d[l[i]]
            del d[l[i]]
            l[i] = obfs_l1
            bdi.append(1)
            y += 1

def group_freqs_result(items):
    items.sort(key=lambda x: x['Word'][0]['n'])
    i = 1
    prev_w = items[0]['Word'][0]['n']
    items[0]['Subitems'] = []
    while i < len(items):
        w = items[i]['Word'][0]['n']
        if w == prev_w:
            items[i-1]['Subitems'].append(items[i])
            del items[i]
        else:
            prev_w = w
            items[i]['Subitems'] = []
            i += 1
    items.sort(key=lambda x: -x['freq'])
    newitems = []
    for item in items:
        newitems.append(item)
        for subitem in item['Subitems']:
            newitems.append(subitem)
        del item['Subitems']
    return newitems


class ConcError (Exception):
    def __init__ (self, msg):
        self.message = msg
    def __str__ (self):
        return self.message


class ConcCGI (UserCGI):

    _global_vars = ['corpname', 'viewmode', 'attrs', 'attr_allpos', 'ctxattrs',
                    'structs', 'refs', 'lemma', 'lpos', 'pagesize',
                    'usesubcorp', 'align', 'gdex_enabled',
                    'gdexcnt', 'gdexconf', 'show_gdex_scores', 'iquery',
                    'maincorp', 'complement_subc', 'attr_tooltip']
    error = ''
    fc_lemword_window_type = 'both'
    fc_lemword_type = 'all'
    fc_lemword_wsize=5
    fc_lemword=''
    fc_pos_window_type = 'both'
    fc_pos_type = 'all'
    fc_pos_wsize=5
    fc_pos=[]
    ml = 0
    concarf = ''
    Aligned = []
    concsize = ''
    samplesize = 10000000 #10M
    fromp = '1'
    fromc = ''
    ml1icase = ''
    ml2icase = ''
    ml3icase = ''
    ml4icase = ''
    ml1bward = ''
    ml2bward = ''
    ml3bward = ''
    freq_sort = ''
    heading = 0
    saveformat = 'text'
    wlattr = ''
    wlpat = ''
    wlpage = 1
    wlcache = ''
    blcache = ''
    simple_n = 1.0
    usearf = 0
    collpage = 1
    fpage = 1
    fmaxitems = 50
    ftt_include_empty = ''
    subcsize = 0
    processing = 0 
    ref_usesubcorp = ''
    wlsort = ''
    keywords = ''
    Keywords = []
    ref_corpname = ''
    Items = []
    selected = ''
    saved_attrs = 0
                     # save options
    pages = 0
    leftctx = ''
    rightctx = ''
    numbering = 0
    align_kwic = 0
    righttoleft = False
    stored = ''
    # end

    corpname = 'susanne'
    corplist = ['susanne', 'bnc']
    usesubcorp = ''
    wlusesubcorp = ''
    obsolete_subcorp = ''
    obsolete_has_subcdef = False
    subcname = ''
    subcpath = []
    _conc_dir = ''
    annotation_group = ''
    _home_url = '../index.html'
    files_path = '..'
    css_prefix = ''
    logo_prefix = ''
    iquery = ''
    queryselector = 'iqueryrow'
    lemma = ''
    lpos = ''
    phrase = ''
    char = ''
    word = ''
    wpos = ''
    cql = ''
    default_attr = None
    save = 1
    isasync = 1
    qmcase = 0
    rlines = '250'
    attrs = 'word'
    ctxattrs = 'word'
    attr_allpos = 'kw'
    allpos = 'kw'
    attr_tooltip = 'nott'
    structs = 'p,g,err,corr'
    q = []
    pagesize = 20
    gdexconf = ''
    gdexpath = [] # [('confname', '/path/to/gdex.conf'), ...]
    gdexcnt = 0
    gdex_enabled = 0
    show_gdex_scores = 0
    alt_gdexconf = None
    _avail_tbl_templates = ''
    multiple_copy = 0
    use_noflash = 0
    select_lines = 0
    line_numbers = 0
    show_permalink = 1
    embed = 0 # TODO: add to globals
    ske_logo = 0
    show_only_first_refs = 0 # TODO: add to globals and to view options
    disable_show_detail = 0
    show_kwic_translations = 1
    wlsendmail = ''
    cup_hl = 'q'

    sortlevel=1
    flimit = 0
    freqlevel=1
    ml1pos = 1
    ml2pos = 1
    ml3pos = 1
    ml4pos = 1
    ml1ctx = '0~0>0'
    ml2ctx = '0~0>0'
    ml3ctx = '0~0>0'
    ml4ctx = '0~0>0'
    tbl_template = 'none'
    errcodes_link = ''
    hidenone = 1
    shorten_refs = 0
    err_types_select = False
    _majorversionerr = []
    _minorversionerr = []
    _version = version.version

    # annotations
    can_annotate = 0
    enable_sadd = 0
    annotconc = ''
    # TODO: put it into run.cgi
    annotconc_init_labels = ('1', '1.a', '1.f', '1.s', 'x', 'u')
    annotconc_num_label_suffixes = ('.a','.f','.a')
    #annotconc_info_umask = 022
    annotconc_info_umask = 0o111  #XXX renumbering from CPA editor
    alpha_features = 0

    # trends
    trends_re = ''
    trends_maxp = 0.1
    trends_method = 'mkts_all'
    trends_sort_by = 't'
    trends_minfreq = 5
    trends_order = 'desc'
    trends_attr = 'word'
    trends_max_items = 1000
    filter_by_trend = 0
    filter_nonwords = 1
    filter_capitalized = 1

    def __init__ (self):
        self.cm = corplib.CorpusManager (self.corplist, self.subcpath, 
                                         self.gdexpath, self._job,
                                         abs_corpname=self.abs_corpname)
        self._curr_corpus = None
        if self._request_method == 'POST':
            self.show_permalink = 0

    def corpora(self):
        l = []
        defs = {
            'id': None,
            'owner_id': None,
            'owner_name': None,
            'tagset_id': None,
            'sketch_grammar_id': None,
            'term_grammar_id': None,
            '_is_sgdev': False,
            'is_featured': False,
            'access_on_demand': False,
            'terms_of_use': None,
            'sort_to_end': None,
            'tags': [],
            'created': None,
            'needs_recompiling': False,
            'user_can_read': True,
            'user_can_upload': False,
            'user_can_manage': False,
            'is_shared': False,
            'is_error_corpus': False
        }
        for c in self.corplist:
            o = dict(defs)
            try:
                mc = conclib.manatee.Corpus(self.abs_corpname(c))
                o.update({
                    'corpname': c,
                    'language_id': mc.get_conf('LANGUAGE'),
                    'language_name': mc.get_conf('LANGUAGE'),
                    'sizes': dict([x.split() for x in mc.get_sizes().split('\n')\
                            if len(x.split(' ')) == 2]),
                    'compilation_status': mc.get_sizes() and 'COMPILED' or 'READY',
                    'new_version': mc.get_conf('NEWVERSION'),
                    'name': mc.get_conf('NAME'),
                    'info': mc.get_conf('INFO'),
                    'wsdef': mc.get_conf('WSDEF'),
                    'termdef': mc.get_conf('TERMDEF'),
                    'diachronic': bool(mc.get_conf('DIACHRONIC')),
                    'aligned': bool(mc.get_conf('ALIGNED')),
                    'docstructure': mc.get_conf('DOCSTRUCTURE')
                })
            except corplib.manatee.CorpInfoNotFound as e:
                o['corpname'] = c
            l.append(o)
        return {'data': l}

    def languages(self):
        langs = set()
        for c in self.corplist:
            try:
                mc = conclib.manatee.Corpus(self.abs_corpname(c))
                lang = mc.get_conf('LANGUAGE')
            except corplib.manatee.CorpInfoNotFound:
                continue
            langs.add(lang)
        data = []
        for lang in langs:
            data.append({
                'id': lang,
                'name': lang,
                'default_tagset_id': None,
                'reference_corpus': None,
                'term_reference_corpus': None,
                'has_term_grammar': False,
            })
        return {'data': data}

    def abs_corpname(self, corpname):
        return corpname

    def now(self):
        return time.strftime('%Y-%m-%d_%H:%M:%S')

    def preprocess_values(self, form):
        cn, rcn, usr, api_key = '', '', '', ''
        if 'json' in form:
            import json
            jsondata = json.loads(form.getvalue('json'))
            cn = str(jsondata.get('corpname', ''))
            rcn = str(jsondata.get('ref_corpname', ''))
            api_key = str(jsondata.get('api_key', ''))
        if 'corpname' in form and not cn:
            cn = form.getvalue('corpname')
        if 'ref_corpname' in form and not rcn:
            rcn = form.getvalue('ref_corpname')
        if 'api_key' in form and not api_key:
            api_key = form.getvalue('api_key')
        if 'username' in form and not usr:
            usr = form.getvalue('username')
        if cn:
            if isinstance(cn, list):
                cn = cn[-1]
            self.corpname = cn
        if rcn: self.ref_corpname = rcn
        self.api_key = api_key
        self.url_username = usr

    def self_encoding(self):
        return (not self._curr_corpus is None) and self._curr_corpus.get_conf('ENCODING') or\
                (self._corp().get_conf('ENCODING') or "utf-8")

    def _corp (self):
        if (not self._curr_corpus or
            (self.usesubcorp and not hasattr(self._curr_corpus, 'subcname'))):
            self._curr_corpus = self.cm.get_Corpus (self.corpname,
                                         self.usesubcorp, self.complement_subc)
            if self.cm.obsolete_subcorp:
                self.obsolete_subcorp = self.cm.obsolete_subcorp
                self.obsolete_has_subcdef = self.cm.obsolete_has_subcdef
                if self.obsolete_has_subcdef:
                    self.rebuild_subc(self.usesubcorp)
            if self.cm.missing_subc_error:
                self.usesubcorp = ''
                raise RuntimeError(self.cm.missing_subc_error)
            # XXX opravit poradne!
            self._curr_corpus.annotation_group = self.annotation_group
            if self.annotation_group:
                self._curr_corpus._conc_dir = os.path.join(self._data_dir, 'skema',
                        self.annotation_group, self.corpname)
            else:
                self._curr_corpus._conc_dir = self._conc_dir
        if self._user and self.user_gdex_path:
            self.cm.user_gdex_path = self.user_gdex_path % self._user
        return self._curr_corpus

    def desc(self, conc=None):
        out = []
        conc_desc = conclib.get_conc_desc (self.q,
                          corpname=self.corpname, _cache_dir=self._cache_dir,
                          subchash=getattr(self._corp(), "subchash", None))
        for o,a,u1,u2,s,f,da in conc_desc:
            rel = ''
            if conc and s != '':
                if float(s) == float(-1):
                    s = conc.fullsize()
                csize = self._corp().search_size()
                rel = round(float(s) * 1000000.0 / csize, 2)
            out.append({'op': o, 'arg': a, 'nicearg': nicearg(a),
                                'rel': rel, 'size': s,
                                'tourl': self.urlencode(u2)})
        return out


    kwicleftctx = '40#'
    kwicrightctx = '40#'
    senleftctx = '-1:s'
    senrightctx = '1:s'
    viewmode = 'kwic'
    align = ''
    sel_aligned = []
    maincorp = ''
    refs_up = 0

    def view_examples(self):
        return self.view()

    def set_default_refs (self):
        if 'refs' not in self.__dict__:
            if self._corp().get_conf('SHORTREF'):
                self.refs = self._corp().get_conf('SHORTREF')
            else:
                sal = self._corp().get_conf('STRUCTATTRLIST').split(',')
                docs = self._corp().get_conf('DOCSTRUCTURE')
                if docs + '.urldomain' in sal:
                    self.refs = '=' + docs + '.urldomain'
                else:
                    self.refs = '#'
        if 'reflinks' not in self.__dict__:
            sal = self._corp().get_conf('STRUCTATTRLIST').split(',')
            c = self._corp()
            for sa in sal:
                if c.get_conf(sa + '.URLTEMPLATE'):
                    self.reflinks.append(sa)

    reflinks = []
    concordance_query = []

    def process_concordance_query(self):
        for obj in self.concordance_query:
            if 'q' in obj:
                if obj['q'].startswith('e') and self.show_gdex_scores:
                    self.q.append('E' + obj['q'][1:])
                else:
                    self.q.append(obj['q'])
            else:
                correct_types(obj, self.defaults)
                for k, v in list(obj.items()): # set attributes
                    setattr(self, k, v)
                if 'pnfilter' in obj: # filter
                    self.q.append(self.filter(return_q=True))
                elif 'sattr' in obj: # simple sort
                    self.q.append(self.sortx(return_q=True))
                elif 'mlsort_options' in obj: # multilevel sort
                    self.q.append(multilevel_sort_crit(obj['mlsort_options']))
                else: # first form
                    self.set_first_query()
                    if obj.get('random', 0):
                        self.q[0] = 'R' + self.q[0]
                for k, v in list(obj.items()): # clean up
                    if k in self.defaults: setattr(self, k, self.defaults[k])
                    else: delattr(self, k)

    def reformat_conc_line(self, line):
        for k in ['Left', 'Kwic', 'Right']:
            split = (k == 'Kwic' and not ',' in self.attrs) or \
                    (k != 'Kwic' and not ',' in self.ctxattrs)
            newsegments = []
            for i, segment in enumerate(line[k]):
                newsegment = {}
                if segment['class'] == 'strc':
                    newsegment['strc'] = segment['str']
                elif segment['class'] == 'coll':
                    newsegment['coll'] = 1
                    newsegment['str'] = segment['str']
                elif segment['class'] == 'attr' and i>0:
                    newsegments[-1]['attr'] = segment['str']
                elif '#' in segment['class']:
                    c = [x for x in segment['class'].split() if x.startswith('#')][0]
                    newsegment['color'] = c
                    newsegment['str'] = segment['str']
                    rem = segment['class'].replace(c, '').strip()
                    if rem:
                        newsegment['class'] = rem
                elif 'conc' in segment['class']:
                    c = [x for x in segment['class'].split() if x.startswith('conc')][0]
                    newsegment['color'] = c[4:]
                    newsegment['str'] = segment['str']
                    rem = segment['class'].replace(c, '').strip()
                    if rem:
                        newsegment['class'] = rem
                else:
                    newsegment['str'] = segment['str']
                if newsegment and split:
                    if 'str' in newsegment: str_c = 'str'
                    else: str_c = 'strc'
                    keys = [x for x in list(newsegment.keys())
                              if x not in ('str', 'strc')]
                    parts = [{str_c: x} for x in newsegment[str_c].split()]
                    for part in parts:
                        for key in keys:
                            part[key] = newsegment[key]
                    newsegments += parts
                elif newsegment:
                        newsegments.append(newsegment)
            line[k] = newsegments

    def concordance (self):
        "concordance API for Platypus"
        self.format = self.format or 'json'
        self.process_concordance_query()
        self.gdex_enabled = 0 # fix for Crystal, might be removed later
        out = self.view()
        if 'error' in out:
            out.update({'concsize': 0, 'fullsize': 0, 'Lines': []})
            return out
        for line in out['Lines']: # reformat the output
            self.reformat_conc_line(line)
            for aligned_line in line.get('Align', []):
                self.reformat_conc_line(aligned_line)
        return out

    def view (self):
        "kwic view"
        self.set_default_refs()
        if self._corp().get_conf('RIGHTTOLEFT') == '1':
            self.righttoleft = True
        if self.viewmode == 'kwic':
            self.leftctx = self.kwicleftctx
            self.rightctx = self.kwicrightctx
        elif self.viewmode == 'align' and self.align:
            self.leftctx = 'a,%s' % os.path.basename(self.corpname)
            self.rightctx = 'a,%s' % os.path.basename(self.corpname)
        else:
            self.leftctx = self.senleftctx
            self.rightctx = self.senrightctx
        # simplified querying of parallel corpora
        if getattr(self, 'l1q', False):
            l1q = getattr(self, 'l1q', '').strip().lower().split()
            l2q = getattr(self, 'l2q', '').strip().lower().split()
            c1 = getattr(self, 'c1', self.corpname)
            c2 = getattr(self, 'c2', '')
            if ''.join(l1q):
                self.q = ['q[lc="' + '"][lc="'.join(l1q) + '"]']
                if ''.join(l2q):
                    self.q[0] += ' within %s: ' % c2
                    self.q[0] += '[lc="' + '"][lc="'.join(l2q) + '"]'
            elif ''.join(l2q):
                self.q = ['q[lc="' + '"][lc="'.join(l2q) + '"]']
                c1, c2 = c2, c1
                self.align = c1
                self.maincorp = c2
            self.q.append('X%s' % c1)

        # GDEX roundabouts
        if self.show_gdex_scores: op = 'E'
        else: op = 'e'
        if self.gdex_enabled and self.gdexcnt: # backward compatibility
            self.q.append('%s%s %s' % (op, str(self.gdexcnt), self.gdexconf))
        for ii, qq in enumerate(self.q): # API use
            if qq.startswith('s*'):
                self.q[ii] = op + qq[2:]

        conc = self.call_function (conclib.get_conc, (self._corp(),))
        if conc is None:
            return {'Lines': [], 'finished': 1, 'concsize': 0, 'fullsize': 0,
                    'isasync': 0, 'Desc': self.desc(conc)}
        conc.switch_aligned (os.path.basename(self.corpname))
        labelmap = {}
        concordance_size_limit = 10 * self._wordlist_max_size
        if self.annotconc:
            if self._selectstored (self.annotconc):
                annot = self._get_annotconc()
                conc.set_linegroup_from_conc (annot)
                labelmap = annot.labelmap
            else:
                self.annotconc = ''
        elif self._wordlist_max_size > 0:
            self.pagesize = min(self.pagesize, concordance_size_limit)
        alignlist = [self.cm.get_Corpus(c) for c in self.align.split(',') if c]
        out = self.call_function(conclib.kwicpage, (conc,), labelmap=labelmap,
                                 pagesize=self.pagesize,
                                 alignlist=alignlist,
                                 tbl_template=self.tbl_template,
                                 reflinks=self.reflinks)
        out['concordance_size_limit'] = concordance_size_limit
        out['Sort_idx'] = self.call_function (conclib.get_sort_idx, (conc,),
                                                      enc=self.self_encoding())
        out['righttoleft'] = self.righttoleft
        out['Aligned_rtl'] = [x.get_conf('RIGHTTOLEFT') == "1" for x in alignlist]
        out['numofcolls'] = conc.numofcolls()
        # get sizes
        out.update({'concsize': not self.error and conc.size() or 0,
                    'finished': int(conc.finished()),
                    'fullsize': conc.fullsize(),
                    'relsize': round(float(1000000.0 * conc.fullsize() / conc.corp().search_size()))})
        if self._corp().get_conf ('ALIGNED'):
            out['Aligned'] = [{'n': w,
                               'label': conclib.manatee.Corpus(w).get_conf(
                                                                  'NAME') or w }
                         for w in self._corp().get_conf ('ALIGNED').split(',')]
        if self.align and not self.maincorp:
            self.maincorp = os.path.basename(self.corpname)
        out['q'] = self.q
        out['Desc'] = self.desc(conc)
        out['port'] = conc.port
        out['gdex_scores'] = conc.gdex_scores
        out['sc_strcts'] = [(st, self._corp().get_conf(st + '.LABEL') or st)
                for st in self._corp().get_conf('STRUCTLIST').split(',') if st]
        return out


    def struct_attr_values(self, struct, attr):
        try:
            a = self._corp().get_struct(struct).get_attr(attr)
            return sorted([a.id2str(i) for i in range(a.id_range()) if a.freq(i) > 0])
        except:
            return []


    def get_conc_sizes (self, q=[], port=0):
        self._headers['Content-Type']= 'text/plain'
        cs = self.call_function (conclib.get_conc_sizes, (self._corp(),),
                                 q=q or self.q, server_port=port)
        return "\n".join(map(str,[cs["finished"], cs["concsize"],
                                  cs["relconcsize"], cs["fullsize"]]))

    def concdesc (self):
        return {'Desc': [{'op': o, 'arg': a, 'churl': self.urlencode(u1), 'da': da,
                          'tourl': self.urlencode(u2), 'size': s, 'formname': f}
                         for o, a, u1, u2, s, f, da in
                         conclib.get_conc_desc (self.q,
                                                corpname=self.corpname,
                                                _cache_dir=self._cache_dir,
                                                subchash=getattr(self._corp(),
                                                             "subchash", None))]
                }
        
    def set_new_viewattrs (self, setattrs=[], allpos='', setstructs=[],
                    setrefs=[], newctxsize='', gdexcnt=0, gdexconf='',
                    attr_tooltip='nott'):
        self.attrs = ','.join(setattrs) or 'word'
        self.structs = ','.join(setstructs)
        self.refs = ','.join(setrefs)
        self.attr_allpos = allpos
        self.attr_tooltip = attr_tooltip
        if allpos == 'all':
            self.ctxattrs = self.attrs
        else:
            self.ctxattrs = 'word'
        self.gdexcnt = gdexcnt
        self.gdexconf = gdexconf
        if newctxsize != self.kwicleftctx[:-1]:
            self.kwicleftctx = newctxsize + '#'
            self.kwicrightctx = newctxsize + '#'

    # sortx options
    sattr = 'word'
    spos = 3
    skey = 'rc'
    sicase = ''
    sbward = ''
    def sortx (self, return_q=False):
        "simple sort concordance"
        if self.skey == 'lc':
            ctx = '-1<0~-%i<0' % self.spos
        elif self.skey == 'kw':
            ctx = '0<0~0>0'
        elif self.skey == 'rc':
            ctx = '1>0~%i>0' % self.spos
        if '.' in self.sattr:
            ctx = ctx.split('~')[0]
        lastq = 's%s/%s%s %s' % (self.sattr, self.sicase, self.sbward, ctx)
        if return_q:
            return lastq
        self.q.append (lastq)
        return self.view()
    
    def mlsortx (self,
          ml1attr='word', ml1pos=1, ml1icase='', ml1bward='', ml1fcode='rc',
          ml2attr='word', ml2pos=1, ml2icase='', ml2bward='', ml2fcode='rc',
          ml3attr='word', ml3pos=1, ml3icase='', ml3bward='', ml3fcode='rc',
          sortlevel=1, ml1ctx='', ml2ctx='', ml3ctx=''):
        "multiple level sort concordance"
        # obsolete, replaced with multilevel_sort for Platypus

        crit = onelevelcrit ('s', ml1attr, ml1ctx, ml1pos, ml1fcode,
                             ml1icase, ml1bward)
        if sortlevel > 1:
            crit += onelevelcrit (' ', ml2attr, ml2ctx, ml2pos, ml2fcode,
                                  ml2icase, ml2bward)
            if sortlevel > 2:
                crit += onelevelcrit (' ', ml3attr, ml3ctx, ml3pos, ml3fcode,
                                      ml3icase, ml3bward)
        self.q.append (crit)
        return self.view()

    def is_err_corpus(self):
        availstruct = self._corp().get_conf('STRUCTLIST').split(',')
        if not ('err' in availstruct and 'corr' in availstruct):
            return False
        return True

    def _compile_basic_query (self, qtype=None, suff='', cname=''):
        queryselector = getattr(self, 'queryselector' + suff)
        iquery = ''
        if queryselector == 'iqueryrow':
            iquery = getattr(self, 'iquery' + suff, '').strip()
            if iquery == '\\':
                iquery = '\\\\'
            iquery = re.sub(r'\\([^.?\\*])', r'\1', iquery)
            iquery = iquery.replace('\\.', '\v')
            iquery = iquery.replace('.', '\\.').replace('\v', '\\.')
            iquery = iquery.replace('\\?', '\v').replace('?', '.')
            iquery = iquery.replace('\v', '?').replace('\\*', '\v')
            iquery = iquery.replace('*', '.*').replace('\v', '\\*')
        lemma = getattr(self, 'lemma' + suff, '')
        lpos = getattr(self, 'lpos' + suff, '')
        phrase = getattr(self, 'phrase' + suff, '')
        qmcase = getattr(self, 'qmcase' + suff, '')
        word = getattr(self, 'word' + suff, '')
        wpos = getattr(self, 'wpos' + suff, '')
        char = getattr(self, 'char' + suff, '')
        cql = getattr(self, 'cql' + suff, '')
        queries = {
            'cql': '%(cql)s',
            'lemma': '[lempos="%(lemma)s%(lpos)s"]',
            'wordform': '[%(wordattr)s="%(word)s" & tag="%(wpos)s.*"]',
            'wordformonly': '[%(wordattr)s="%(word)s"]',
            }
        for a in ('iquery', 'word', 'lemma', 'phrase', 'cql', 'char'):
            if queryselector == a + 'row':
                if getattr(self, a+suff, ''):
                    setattr (self, a+suff, getattr (self, a+suff).strip())
                elif suff:
                    return ''
                else:
                    raise ConcError('No query entered.')
        if qtype:
            return queries[qtype] % self.clone_self()
        thecorp = cname and self.cm.get_Corpus (cname) or self._corp()
        attrlist = thecorp.get_conf('ATTRLIST').split(',')
        wposlist = dict (corpconf_pairs (thecorp, 'WPOSLIST'))
        lposlist = dict (corpconf_pairs (thecorp, 'LPOSLIST'))
        if queryselector == 'iqueryrow':
            qitem = thecorp.get_conf('SIMPLEQUERY').replace('%s', '%(q)s')
            if '--' not in iquery:
                iqresult = ''
                for iqpart in iquery.split('|'):
                    if iqresult:
                        iqresult += ' | '
                    iqresult += ''.join([qitem % {'q':escape_nonwild(q)}
                                         for q in iqpart.split()])
                return iqresult
            else:
                def split_tridash (word, qitem):
                    if '--' not in word:
                          return qitem % {'q':word}
                    w1,w2 = word.split('--',1)
                    return "( %s | %s %s | %s )" % (qitem % {'q':w1+w2}, 
                                                    qitem % {'q':w1}, 
                                                    qitem % {'q':w2}, 
                                                    qitem % {'q':w1+'-'+w2})
                return ''.join([split_tridash(escape_nonwild(q), qitem)
                                for q in iquery.split()])

        if queryselector == 'lemmarow':
            if not lpos:
                if qmcase:
                    return '[lemma="%s"]' % lemma
                else:
                    if 'lemma_lc' in attrlist:
                        return '[lemma_lc="%s"]' % lemma
                    else:
                        return '[lemma="(?i)%s"]' % lemma
            if 'lempos' in attrlist:
                try:
                    if not lpos in list(lposlist.values()):
                        lpos = lposlist [lpos]
                except KeyError:
                    raise ConcError ('Undefined lemma PoS' + ' "%s"' % lpos)
                if qmcase:
                    return '[lempos="(%s)%s"]' % (lemma, lpos)
                else:
                    if 'lempos_lc' in attrlist:
                        return '[lempos_lc="(%s)%s"]' % (lemma, lpos)
                    else:
                        return '[lempos="(?i)(%s)%s"]' % (lemma, lpos)
            else: # XXX
                try:
                    if lpos in list(wposlist.values()):
                        wpos = lpos
                    else:
                        wpos = wposlist [lpos]
                except KeyError:
                    raise ConcError ('Undefined word form PoS' + ' "%s"' % lpos)
                if qmcase:
                    return '[lemma="%s" & tag="%s"]' % (lemma, wpos)
                else:
                    if 'lemma_lc' in attrlist:
                        return '[lemma_lc="%s" & tag="%s"]' % (lemma, wpos)
                    else:
                        return '[lemma="(?i)%s" & tag="%s"]' % (lemma, wpos)
        if queryselector == 'phraserow':
            if qmcase:
                return ' | '.join(['"' + '" "'.join(phr.split()) + '"'
                        for phr in phrase.split('|')])
            else:
                if "lc" in attrlist:
                    return ' | '.join([" ".join(['[lc="%s"]' % x for x in phr.split()])
                            for phr in phrase.split('|')])
                else:
                    return ' | '.join([" ".join(['[word="(?i)%s"]' % x for x in phr.split()])
                            for phr in phrase.split('|')])
        if queryselector == 'wordrow':
            if qmcase:
                wordattr = 'word="%s"' % word
            else:
                if 'lc' in attrlist:
                    wordattr = 'lc="%s"' % word
                else:
                    wordattr = 'word="(?i)%s"' % word
            if not wpos:
                return '[%s]' % wordattr
            try:
                if not wpos in list(wposlist.values()):
                    wpos = wposlist [wpos]
            except KeyError:
                raise ConcError ('Undefined word form PoS' + ' "%s"' % wpos)
            return '[%s & tag="%s"]' % (wordattr, wpos)
        if queryselector == 'charrow':
            if not char: raise ConcError('No char entered')
            return '[word=".*%s.*"]' % escape(char)
        return cql


    def _compile_query(self, qtype=None, cname=''):
        if not self.is_err_corpus():
            return self._compile_basic_query(qtype, cname=cname)
        err_code = getattr(self, 'cup_err_code', '')
        err = getattr(self, 'cup_err', '')
        corr = getattr(self, 'cup_corr', '')
        switch = getattr(self, 'errcorr_switch', '')
        if not err_code and not err and not corr:
            cql = self._compile_basic_query(qtype)
            if self.queryselector != 'cqlrow':
                cql = cql.replace('][', '] (<corr/>)? [')
                cql = cql.replace('](', '] (<corr/>)? (')
                cql = cql.replace('] [', '] (<corr/>)? [')
            return cql
        # compute error query
        # Achtung: Top error lists (method freq) depends on this -- do check it
        # when changing the below
        corr_restr = corr or (err_code and switch == 'c')
        err_restr = err or (err_code and switch == 'e')
        if err_code: corr_within = '<corr type="%s"/>' % err_code
        else: corr_within = '<corr/>'
        if err_code: err_within = '<err type="%s"/>' % err_code
        else: err_within = '<err/>'
        err_containing = ''; corr_containing = ''
        if err:
            self.iquery = err; self.queryselector = 'iqueryrow'
            err_containing = ' containing ' + self._compile_basic_query(qtype)
        if corr:
            self.iquery = corr; self.queryselector = 'iqueryrow'
            corr_containing = ' containing ' + self._compile_basic_query(qtype)
        err_query =  '(%s%s)' % (err_within, err_containing)
        corr_query = '(%s%s)' % (corr_within, corr_containing)
        fullstruct = '(%s%s)' % (err_query, corr_query)
        if self.cup_hl == 'e' or (self.cup_hl == 'q' and err_restr
                                                     and not corr_restr):
            return '%s within %s' % (err_query, fullstruct)
        elif self.cup_hl == 'c' or (self.cup_hl == 'q' and corr_restr
                                                       and not err_restr):
            return '%s within %s' % (corr_query, fullstruct)
        else: # highlight both
            return fullstruct

    def query (self, qtype='cql'):
        "perform query"
        if self.default_attr:
            qbase = 'a%s,' % self.default_attr 
        else:
            qbase = 'q'
        self.q = [qbase + self._compile_query()]
        return self.view()

    def set_first_query (self):
        def append_filter (attrname, items, ctx, fctxtype):
            if not items:
                return
            if fctxtype == 'any':
                self.q.append ('P%s [%s]' %
                               (ctx, '|'.join (['%s="%s"' % (attrname, i)
                                                for i in items])))
            elif fctxtype == 'none':
                self.q.append ('N%s [%s]' %
                          (ctx, '|'.join (['%s="%s"' % (attrname, i)
                                           for i in items])))
            elif fctxtype == 'all':
                for i in items:
                    self.q.append ('P%s [%s="%s"]' % (ctx, attrname, i))
                
        if 'lemma' in self._corp().get_conf('ATTRLIST').split(','):
            lemmaattr = 'lemma'
        else:
            lemmaattr = 'word'
        wposlist = dict (corpconf_pairs (self._corp(), 'WPOSLIST'))
        if self.queryselector == 'phraserow':
            self.default_attr = 'word' # XXX to be removed with new first form
        if self.default_attr:
            qbase = 'a%s,' % self.default_attr 
        else:
            qbase = 'q'
        texttypes = self._texttype_query()
        if texttypes:
            ttquery = ' ' + ' '.join (['within <%s %s />' % nq for nq in texttypes])
        else:
            ttquery = ''
        par_query = ''
        nopq = []
        for al_corpname in self.sel_aligned:
            if getattr(self, 'pcq_pos_neg_' + al_corpname) == 'pos': wnot = ''
            else: wnot = '!'
            pq = self._compile_basic_query(suff='_'+al_corpname,
                                           cname=al_corpname)
            if pq: par_query += ' within%s %s:%s' % (wnot, al_corpname, pq)
            if not pq or wnot: nopq.append(al_corpname)
        self.q = [qbase + self._compile_query() + ttquery + par_query]

        if self.fc_lemword_window_type == 'left':
            append_filter (lemmaattr,
                           self.fc_lemword.split(),
                           '-%i -1 -1' % self.fc_lemword_wsize,
                           self.fc_lemword_type)
        elif self.fc_lemword_window_type == 'right':
            append_filter (lemmaattr,
                           self.fc_lemword.split(),
                           '1 %i 1' % self.fc_lemword_wsize,
                           self.fc_lemword_type)
        elif self.fc_lemword_window_type == 'both':
            append_filter (lemmaattr,
                           self.fc_lemword.split(),
                           '-%i %i 1' % (self.fc_lemword_wsize,
                                         self.fc_lemword_wsize),
                           self.fc_lemword_type)
        if self.fc_pos_window_type == 'left':
            append_filter ('tag',
                           [wposlist.get(t,'') for t in self.fc_pos],
                           '-%i -1 -1' % self.fc_pos_wsize,
                           self.fc_pos_type)
        elif self.fc_pos_window_type == 'right':
            append_filter ('tag',
                           [wposlist.get(t,'') for t in self.fc_pos],
                           '1 %i 1' % self.fc_pos_wsize,
                           self.fc_pos_type)
        elif self.fc_pos_window_type == 'both':
            append_filter ('tag',
                           [wposlist.get(t,'') for t in self.fc_pos],
                           '-%i %i 1' % (self.fc_pos_wsize, self.fc_pos_wsize),
                           self.fc_pos_type)
        for al_corpname in self.sel_aligned:
            if al_corpname in nopq and getattr(self,
                                         'filter_nonempty_' + al_corpname, ''):
                self.q.append('X%s' % al_corpname)
        if self.sel_aligned: self.align = ','.join(self.sel_aligned)


    def first (self):
        self.set_first_query()
        return self.view()
    
    # filter settings
    pnfilter = 'p'
    filfl = 'f'
    filfpos = '-5'
    filtpos = '5'
    within = 0
    inclkwic = False

    def filter (self, return_q=False):
        "Positive/Negative filter"
        if not self.inclkwic:
            self.pnfilter = self.pnfilter.upper()
        rank = {'f':1, 'l':-1}.get (self.filfl, 1)
        texttypes = self._texttype_query()
        try:
            query = self._compile_query(cname=self.maincorp)
        except ConcError:
            if texttypes: query = '[]'; self.filfpos='0'; self.filtpos='0'
            else: raise ConcError ('No query entered.')
        query +=  ' '.join (['within <%s %s />' % nq for nq in texttypes])
        if self.within:
            kw = self.pnfilter in 'nN' and ' within! ' or ' within '
            wquery =  kw + '%s:(%s)' % (self.maincorp or self.corpname, query)
            self.q[0] += wquery
            lastq = 'x-' + (self.maincorp or self.corpname)
        else:
            lastq = '%s%s %s %i %s' % (self.pnfilter, self.filfpos,
                                       self.filtpos, rank, query)
        if return_q:
            return lastq
        self.q.append(lastq)
        try:
            return self.view()
        except:
            if self.within: self.q[0] = self.q[0][:-len(wquery)]
            else: del self.q[-1]
            raise


    def reduce (self, rlines='250'):
        "random sample"
        self.q.append ('r' + rlines)
        return self.view()
       
    fcrit = []
    examples = 0
    group = 0
    hitlen = 0
    def freqs(self, fcrit=[], flimit=0, freq_sort='', ml=0):
        "display a frequency list"
        import operator
        def parse_fcrit(fcrit):
            attrs, marks, ranges = [], [], []
            for i, item in enumerate(fcrit.split()):
                if i % 2 == 0: attrs.append(item)
                if i % 2 == 1: ranges.append(item)
            return attrs, ranges
        if self._wordlist_max_size:
            if not self.fmaxitems:
                self.fmaxitems = self._wordlist_max_size
            else:
                self.fmaxitems = min(self.fmaxitems, self._wordlist_max_size)
        elif not self.fmaxitems:
            self.fmaxitems = 10000000
        corp = self._corp()
        csize = corp.size()
        if self.concordance_query: self.process_concordance_query()
        conc = self.call_function (conclib.get_conc, (corp, self.samplesize))
        conc.sync()
        if not self.hitlen:
            kwcls = self.call_function(conclib.kwicpage, (conc,))['Lines']
            self.hitlen = max([line['hitlen'] for line in kwcls])
        result = {'fcrit': self.urlencode ([('fcrit', self.rec_recode(cr))
                                            for cr in fcrit]),
                  'FCrit': [{'fcrit': cr} for cr in fcrit],
                  'Blocks': [conc.xfreq_dist (cr, flimit, freq_sort, 300, ml,
                                   self.ftt_include_empty, self.wlmaxfreq) for cr in fcrit],
                  'paging': 0,
                  'concsize':conc.size(),
                  'fullsize':conc.fullsize(),
                  'Desc': self.desc(conc),
                  'numofcolls': conc.numofcolls(),
                  'hitlen': self.hitlen,
                  'wllimit': self._wordlist_max_size }
        if not result['Blocks'][0]: raise ConcError('Empty list')
        if len(result['Blocks']) == 1 and self.group:
            result['Blocks'][0]['Items'] = group_freqs_result(
                                                    result['Blocks'][0]['Items'])
        if len(result['Blocks']) == 1 and not self.examples: # paging
            fstart = (self.fpage - 1) * self.fmaxitems
            self.fmaxitems = self.fmaxitems * self.fpage + 1
            result['paging'] = 1
            if len(result['Blocks'][0]['Items']) < self.fmaxitems:
                result['lastpage'] = 1
            else:
                result['lastpage'] = 0
            result['Blocks'][0]['Items'] = \
               result['Blocks'][0]['Items'][fstart:self.fmaxitems-1]
        for b in result['Blocks']:
            if self._wordlist_max_size and (self._wordlist_max_size
                                                 < len(b['Items'])):
                b['Items'] = b['Items'][:self._wordlist_max_size]
                result['lastpage'] = 1
                b['note'] = 'You are allowed to see only %s items.' \
                               % self._wordlist_max_size
            for item in b['Items']:
                item['pfilter'], item['nfilter'] = '', ''
                item['pfilter_list'] = []
                item['fpm'] = (item.get('freq', 0) / float(csize)) * 1000000
        ## generating positive and negative filter references
        for b_index, block in enumerate(result['Blocks']):
            curr_fcrit = fcrit[b_index]
            attrs, ranges = parse_fcrit(curr_fcrit)
            for level, (attr, range) in enumerate(zip(attrs, ranges)):
                begin = range.split('~')[0]
                if '~' in range: end = range.split('~')[1]
                else: end = begin
                attr = attr.split("/")
                if len(attr) > 1 and "i" in attr[1]: icase = '(?i)'
                else: icase = ''
                attr = attr[0]
                for ii, item in enumerate(block['Items']):
                    if not item['freq']: continue
                    if not '.' in attr:
                        if attr in corp.get_conf('ATTRLIST').split(','):
                            wwords = item['Word'][level]['n'].split('  ') # two spaces
                            fquery = '%s %s 0 ' % (begin, end)
                            fquery += ''.join(['[%s="%s%s"]'
                                % (attr, icase, escape(w)) for w in wwords ])
                        else: # structure number
                            fquery = '0 0<0 1 [] within <%s #%s/>' % \
                                      (attr, item['Word'][0]['n'].split('#')[1])
                    else: # text types
                        structname, attrname = attr.split('.')
                        if corp.get_conf(structname + '.NESTED'):
                            block['unprecise'] = True
                        fquery = '0 0<0 1 [] within <%s %s="%s" />' \
                                 % (structname, attrname,
                                    escape(item['Word'][level]['n']))
                    if not item['freq']: continue
                    efquery = self.urlencode(fquery)
                    item['pfilter'] += ';q=p%s' % efquery
                    item['pfilter_list'].append('p%s' % fquery)
                    if len(attrs) == 1 and item['freq'] <= conc.size():
                        item['nfilter'] += ';q=n%s' % efquery
        # adding no error, no correction (originally for CUP)
        errs, corrs, err_block, corr_block = 0, 0, -1, -1
        for b_index, block in enumerate(result['Blocks']):
           curr_fcrit = fcrit[b_index]
           if curr_fcrit.split()[0] == 'err.type':
               err_block = b_index
               for item in block['Items']: errs += item['freq']
           elif curr_fcrit.split()[0] == 'corr.type':
               corr_block = b_index
               for item in block['Items']: corrs += item['freq']
        freq = conc.size() - errs - corrs
        if freq > 0 and err_block > -1 and corr_block > -1:
            pfilter = ';q=p0 0<0 1 ([] within ! <err/>) within ! <corr/>'
            cc = self.call_function(conclib.get_conc, (corp,),
                                    q=self.q + [pfilter[3:]])
            freq = cc.size()
            err_nfilter, corr_nfilter = '', ''
            if freq != conc.size():
                err_nfilter = ';q=p0 0<0 1 ([] within <err/>) within ! <corr/>'
                corr_nfilter = ';q=p0 0<0 1 ([] within! <err/>) within <corr/>'
            result['Blocks'][err_block]['Items'].append(
                              {'Word': [{'n': 'no error' }], 'freq': freq,
                               'pfilter': pfilter, 'nfilter': err_nfilter,
                               'norel': 1, 'fbar' :0} )
            result['Blocks'][corr_block]['Items'].append(
                         {'Word': [{'n': 'no correction' }], 'freq': freq,
                          'pfilter': pfilter, 'nfilter': corr_nfilter,
                          'norel': 1, 'fbar' :0} )
        return result

    def freqml (self, flimit=0, freqlevel=1,
                ml1attr='word', ml1pos=1, ml1icase='', ml1fcode='rc',
                ml2attr='word', ml2pos=1, ml2icase='', ml2fcode='rc',
                ml3attr='word', ml3pos=1, ml3icase='', ml3fcode='rc',
                ml4attr='word', ml4pos=1, ml4icase='', ml4fcode='rc',
                ml5attr='word', ml5pos=1, ml5icase='', ml5fcode='rc',
                ml6attr='word', ml6pos=1, ml6icase='', ml6fcode='rc',
                ml1ctx='0', ml2ctx='0', ml3ctx='0', ml4ctx='0', ml5ctx='0',
                ml6ctx='0'):
        "multilevel frequency list"
        l = locals()
        fcrit = ' '.join([onelevelcrit ('', l['ml%dattr'%i],
                                        l['ml%dctx'%i], l['ml%dpos'%i],
                                        l['ml%dfcode'%i], l['ml%dicase'%i], 'e')
                         for i in range(1, freqlevel+1)])
        ml = '.' not in ml1attr
        result = self.freqs([fcrit], flimit, self.freq_sort, ml)
        result['ml'] = ml
        return result

    def freqtt (self, flimit=0, fttattr=[]):
        if not fttattr:
            raise ConcError('No text type selected')
        return self.freqs (['%s 0' % a for a in fttattr], flimit)

    cattr = 'word'
    csortfn = 'd'
    cbgrfns = 'mtd'
    cfromw = -5
    ctow = 5
    cminfreq = 5
    cminbgr = 3
    cmaxitems = 50
    def collx (self, cattr='', csortfn='d', cbgrfns=['t','m', 'd'], usesubcorp=''):
        "list collocations"
        corp = self._corp() # !! must precede manatee.Corpus(maincorp)
        if self.maincorp:
            corp = conclib.manatee.Corpus(self.abs_corpname(self.maincorp))
        if usesubcorp and not corplib.are_subcorp_stats_compiled (corp, cattr):
            out = corplib.compute_freqfile (self._user, corp, cattr,
                                            self.wlnums, self.get_url())
            return { 'processing': out["progress"] == '100' and '99' \
                                   or out["progress"],
                     'bgjob_notification': out['notifyme'] and 'checked' or '',
                     'jobid': out["jobid"], 'esttime': out["esttime"],
                     'new_maxitems': self.wlmaxitems}
        if self._wordlist_max_size:
            if not self.cmaxitems:
                self.cmaxitems = self._wordlist_max_size
            else:
                self.cmaxitems = min(self.cmaxitems, self._wordlist_max_size)
        elif not self.cmaxitems:
            self.cmaxitems = 10000000
        collstart = (self.collpage - 1) * self.cmaxitems
        self.cmaxitems = self.cmaxitems * self.collpage + 1
        self.cbgrfns = ''.join (cbgrfns)
        if not csortfn and cbgrfns:
            self.csortfn = cbgrfns[0]
        if self.concordance_query: self.process_concordance_query()
        conc = self.call_function (conclib.get_conc, (self._corp(),))
        conc.sync()
        result = self.call_function (conc.collocs, ())
        if len(result['Items']) < self.cmaxitems:
            result['lastpage'] = 1
        else:
            result['lastpage'] = 0;
            result['Items'] = result['Items'][:-1]
        result['wllimit'] = self._wordlist_max_size
        if result['wllimit'] and (result['wllimit'] < len(result['Items'])):
                result['Items'] = result['Items'][:self._wordlist_max_size]
                result['lastpage'] = 1
                result['note'] = 'You are allowed to see only %s items.' \
                                 % self._wordlist_max_size
        result['Items'] = result['Items'][collstart:]
        for item in result['Items']:
            item["pfilter"] = 'q=' + self.urlencode(item["pfilter"])
            item["nfilter"] = 'q=' + self.urlencode(item["nfilter"])
        result['concsize'] = conc.size()
        result['Desc'] = self.desc(conc)
        return result

    def structctx (self, pos=0, struct='doc'):
        "display a hit in a context of a structure"
        s = self._corp().get_struct(struct)
        struct_id = s.num_at_pos(pos)
        beg, end = s.beg(struct_id), s.end(struct_id)
        self.detail_left_ctx = pos - beg
        self.detail_right_ctx = end - pos - 1
        result = self.widectx(pos)
        result ['no_display_links'] = True
        return result
        
    def widectx (self, pos=0):
        "display a hit in a wider context"
        return self.call_function (conclib.get_detail_context, (self._corp(),
                                                                pos))

    def fullref (self, pos=0):
        "display a full reference"
        return self.call_function (conclib.get_full_ref, (self._corp(), pos))

    def freq_distrib(self, fcrit='', flimit=0, res=100, ampl=101, format=''):
        "get data for frequency distribution graph"
        self.process_concordance_query()
        self.fcrit = fcrit
        conc = self.call_function(conclib.get_conc, (self._corp(),))
        conc.sync()
        values = conclib.manatee.IntVector([0]*res)
        begs = conclib.manatee.IntVector([0]*res)
        conc.distribution(values, begs, ampl)
        begs2 = [j and conc.beg_at(i) or -1 for i, j in zip(begs, values)]
        ends = [0]*res
        i, j = 0, 1
        while j < res:
            if values[i]:
                if values[j]:
                    ends[i] = begs2[j]-1
                    j += 1
                    i = j-1
                else:
                    j += 1
            else:
                j += 1
                i += 1
        ends[i] = conc.end_at(conc.size()-1)
        return {
                'dots': [{
                    'frq': v[0],
                    'pos': i/float(len(values)),
                    'beg': int(v[1]),
                    'end': int(v[2]),
                    } for i, v in enumerate(zip(values, begs2, ends))],
                'granularity': res
        }

    def clear_cache (self, corpname=''):
        if not corpname: corpname = self.corpname
        if '..' in corpname or corpname.startswith('/'):
            return {'error': 'This action is not allowed.'}
        os.system ('rm -rf %s/%s' % (self._cache_dir, corpname))
        return {'message': 'Cache cleared: %s' % corpname}

    wlminfreq = 5
    wlmaxfreq = 0
    wlmaxitems = 100
    wlwords = []
    blacklist = []
    _wordlist_max_size = 0

    def get_wl_words(self, attrnames=('wlfile', 'wlcache')):
            # gets arbitrary list of words for wordlist
        wlfile = getattr(self, attrnames[0], '')
        wlcache = getattr(self, attrnames[1], '')
        filename = wlcache; wlwords = []
        if wlfile: # save a cache file
            try:
                from hashlib import md5
            except ImportError:
                from md5 import new as md5
            filename = os.path.join(self._cache_dir,
                                    md5(wlfile).hexdigest() + '.wordlist')
            if not os.path.isdir(self._cache_dir): os.makedirs (self._cache_dir)
            cache_file = open(filename, 'w')
            cache_file.write(wlfile)
            cache_file.close()
            wlwords = [w.strip() for w in wlfile.splitlines()]
        if wlcache: # read from a cache file
            filename = os.path.join(self._cache_dir, wlcache)
            cache_file = open(filename)
            wlwords = [w.strip()
                       for w in cache_file.read().splitlines()]
            cache_file.close()
        return wlwords, os.path.basename(filename)


    include_nonwords = 0
    usengrams = 0
    ngrams_n = 2
    ngrams_max_n = 2
    nest_ngrams = 0
    wltype = 'simple'
    wlnums = 'frq'
    complement_subc = 0

    def check_wl_compatibility (self, wltype, ref_subcorp):
        if self.usengrams or ref_subcorp == '== the rest of the corpus ==':
            if self.usengrams and self.wlnums != 'frq':
                raise ConcError('N-grams and "rest of corpus" can use only '
                                  'raw hit counts')
            if wltype == 'multilevel':
                raise ConcError('N-grams are not compatible with changing '
                                  'output attribute(s)')
            if self.wlattr in ('ws_collocations', 'ws_terms') \
                     or '.' in self.wlattr:
                raise ConcError('N-grams and "rest of corpus" can use only '
                                  'positional attributes')
        if self.usengrams and ref_subcorp == '== the rest of the corpus ==':
            raise ConcError('N-grams and "rest of corpus" can not be used'
                              'together')
        if '.' in self.wlattr:
            if wltype != 'simple':
                raise ConcError('Text types are limited to simple output')
            if self.wlnums == 'arf':
                raise ConcError('ARF cannot be used with text types')
            elif self.wlnums == 'frq':
                self.wlnums = 'doc sizes'
            elif self.wlnums == 'docf':
                self.wlnums = 'frq'
        if self.wlattr == 'ws_collocations' and self.wlnums != 'frq':
            raise ConcError('Word sketch collocations are available '
                            'with raw hit counts only')
        if self.wlattr == 'ws_terms' and self.wlnums != 'frq':
            raise ConcError('Terms are available with raw hit counts only')

    def wipo (self, single=0):
        from wipo_filters import filter_wipo_terms
        if self._wordlist_max_size:
            self.wlmaxitems = min(self.wlmaxitems, self._wordlist_max_size)
        self.blacklist, self.blcache = self.get_wl_words(('wlblacklist',
                                                          'blcache'))
        lang = self._corp().get_conf('LANGUAGE').split(', ')[0]
        keyterms, size = corplib.ws_keywords(self._corp(),
                                     corplib.manatee.Corpus(self.abs_corpname(self.ref_corpname)),
                                     wlminfreq=1, wlmaxitems=self.wlmaxitems*5,
                                     simple_n=1, ret_count=True,
                                     wlattr='ws_terms', filter_singleword=0)
        if size < 100000: # count the exact numbers
            keyterms, size = corplib.ws_keywords(self._corp(),
                                     corplib.manatee.Corpus(self.abs_corpname(self.ref_corpname)),
                                     wlminfreq=1, wlmaxitems=100000,
                                     simple_n=1, ret_count=True,
                                     wlattr='ws_terms', filter_singleword=0)
        data = []
        for s, rel, relref, f, fr, w in keyterms:
            sitem = w.split('\t')
            data.append({
                'str': w.replace('_', ' '),
                'score': round(s, 1),
                'frq1': int(f),
                'frq2': int(fr),
                'relfrq1': round(rel, 1),
                'relfrq2': round(relref, 1),
                'w1': sitem[0],
                'gramrel': sitem[1],
                'item': sitem[2].replace('_', ' '),
                'seek': sitem[3],
                'query': '[term(2, %s)]' % sitem[3],
                'ref_seek': sitem[4],
                'ref_query': '[term(2, %s)]' % sitem[4]
            })
        keyterms = [x for x in filter_wipo_terms(data, lang, single)
                          if x['str'].split('\t')[2] not in self.blacklist]
        if size < 100000: size = len(keyterms)
        return { 'Keyterms': keyterms[:self.wlmaxitems],
                 'single': single,
                 'lang': lang,
                 'size': size,
               }


    def do_nest_ngrams(self, items): # items must be sorted by freq
        from operator import itemgetter
        ngrams = [[], [], [], [], [], [], [], []]
             # (0-), (1-), 2-, 3-, 4-, 5-, 6-, (7-)grams
        for item in items:
            n = item['str'].count('  ') + 1
            item['str'] = '  ' + item['str'] + '  '
            item['children'] = []
            item['nested'] = False
            for ng in reversed(ngrams[n+1]): # find superior n+1-gram
                if ng['freq'] > item['freq']:
                    break
                if item['str'] in ng['str']: # ng['freq'] == item['freq']
                    item['nested'] = True
                    ng['children'].append(item)
            for ng in ngrams[n-1]: # find subordinate (n-1)-grams
                if ng['str'] in item['str']:
                    item['children'].append(ng)
                    ng['nested'] = True
            ngrams[n].append(item)
        result = ngrams[2]+ngrams[3]+ngrams[4]+ngrams[5]+ngrams[6]
        if self.wlsort == 'f': result.sort(key=itemgetter('freq'), reverse=True)
        else: result.sort(key=itemgetter('str'))
        for item in result:
            item['str'] = item['str'][2:-2]
        return result


    def wordlist (self, wlpat='', wltype='simple', corpname='', usesubcorp='',
                  ref_corpname='', ref_usesubcorp='', wlpage=1):
        if self.wlusesubcorp: usesubcorp = self.usesubcorp = self.wlusesubcorp
        if not wlpat: self.wlpat = '.*'
        if self.wlpat != '.*':
            m = re.match(r'(\(\?[a-zA-Z]\))?(.*)', self.wlpat)
            self.wlpat = '%s^%s$' % (m.group(1) if m.group(1) else "", m.group(2))
        if '.' in self.wlattr: orig_wlnums = self.wlnums
        self.check_wl_compatibility(wltype, ref_usesubcorp)
        result= { 'new_maxitems': self.wlmaxitems,
                  'wllimit': self._wordlist_max_size }
        wlstart = (wlpage - 1) * self.wlmaxitems
        self.wlmaxitems =  self.wlmaxitems * wlpage + 1 # +1 = end detection
        if self._wordlist_max_size and (self._wordlist_max_size
                                              < self.wlmaxitems):
            self.wlmaxitems = self._wordlist_max_size + 1
            result['lastpage'] = 1
            result['note'] = 'You are allowed to see only %s items.' \
                               % self._wordlist_max_size
        try:
            self.words, self.wlcache = self.get_wl_words()
            self.blacklist, self.blcache = self.get_wl_words(('wlblacklist',
                                                                    'blcache'))
            if wltype == 'keywords':
                sc = self.cm.get_Corpus (self.corpname, usesubcorp,
                                         self.complement_subc)
                if ref_usesubcorp == '== the rest of the corpus ==':
                    scref = self.cm.get_Corpus (ref_corpname, usesubcorp,
                                                complement=True)
                else:
                    scref = self.cm.get_Corpus (ref_corpname, ref_usesubcorp)
                if self.wlattr in ('ws_terms', 'ws_collocations'):
                    kw_func = getattr(corplib, 'ws_keywords')
                elif self.usengrams:
                    kw_func = getattr(corplib, 'ngr_keywords')
                else:
                    kw_func = getattr(corplib, 'subc_keywords_onstr')
                out, total, totalfrq = self.call_function (kw_func, (sc, scref))
                result_list = [{
                    'str': w, 'score': round(s,1),
                    'freq': round(f, 1) if self.wlnums == 'arf' else int(f),
                    'freq_ref': round(fr, 1) if self.wlnums=='arf' else int(fr),
                    'rel': round(rel, 1),
                    'rel_ref': round(relref, 1) }
                       for s, rel, relref, f, fr, w in out]
                if self.wlattr in ('ws_terms', 'ws_collocations'):
                    for item in result_list:
                        sitem = item['str'].split('\t')
                        item['w1'] = sitem[0]
                        item['gramrel'] = sitem[1]
                        item['w2'] = sitem[2]
                        item['seek'] = sitem[3]
                        item['q'] = '[ws(2, %s)]' % sitem[4]
                        item['ref_seek'] = sitem[4]
                result['ref_corp_full_name'] = scref.get_conf('NAME') \
                                               or scref.get_conffile()
                result['total'] = total
                result['totalfrq'] = totalfrq
            else: # ordinary list
                if self.wlattr in ('ws_terms', 'ws_collocations'):
                    result_list = self.call_function (corplib.ws_wordlist,
                                                      (self._corp(),))
                elif self.usengrams:
                    result['corpsize'] = self._corp().size()
                    if self.nest_ngrams:
                        wlstart = 0
                        self.wlmaxitems = max(self.wlmaxitems, self.wlpage*1000)
                        if self._wordlist_max_size \
                              and self._wordlist_max_size < self.wlpage * 1000:
                            self.wlmaxitems = self._wordlist_max_size
                    result_list, cnt, f = self.call_function (corplib.ngrams,
                                                      (self._corp(),))
                    if self.nest_ngrams:
                        if 0 < self._wordlist_max_size <= len(result_list):
                            result['note'] = 'You are allowed to see only %s'\
                                           ' items.' % self._wordlist_max_size
                        result_list = self.do_nest_ngrams(result_list)
                        result['nested'] = self.nest_ngrams
                    result['total'] = cnt
                    result['totalfrq'] = f
                else:
                    try:
                        result_list, cnt, f = self.call_function (corplib.wordlist,
                                                                  (self._corp(),))
                    except RuntimeError:
                        raise corplib.MissingSubCorpFreqFile(self._corp())
                    result['total'] = cnt
                    result['totalfrq'] = f
                    if self.words: result['wlcache'] = self.wlcache
                    if self.blacklist: result['blcache'] = self.blcache
            if len(result_list) < self.wlmaxitems: result['lastpage'] = 1
            else: result['lastpage'] = 0; result_list = result_list[:-1]
            result_list = result_list[wlstart:]
            if self.wlattr not in ('ws_terms', 'ws_collocations'):
                for ll in result_list: # url-encoded queries to concordance
                    encstr = ll['str'].replace('\\','\\\\"').replace('"','\\"')
                    ll['q'] = self.urlencode(
                                   ''.join(['[%s=="%s"]' % (self.wlattr,w)
                                            for w in encstr.split('  ')]))
            if wltype == 'keywords': result['Keywords'] = result_list
            else: result['Items'] = result_list
            self.wlmaxitems -= 1
            if '.' in self.wlattr: self.wlnums = orig_wlnums
            try:
                if not wlattr in ('ws_terms', 'ws_collocations'):
                    result['wlattr_label'] = self._corp().get_conf(
                                           self.wlattr+'.LABEL') or self.wlattr
            except:
                result['wlattr_label'] = self.wlattr
            result['frtp'] = corplib.get_stat_desc (self.wlnums)
            return result
        except corplib.MissingSubCorpFreqFile as subcmiss: # run compilation
            self.wlmaxitems -= 1
            if self.usengrams:
                if type('') == type(subcmiss.args[0]): # missing n-grams
                    out = corplib.compute_ngrams (self._user, subcmiss.args[1],
                                                  subcmiss.args[0],
                                                  self.get_url())
                else: # freqfile miss
                    out = corplib.compute_freqfile (self._user,
                          subcmiss.args[0], self.wlattr + '.ngr',
                          self.wlnums, self.get_url())
            else:
                if self.wlattr not in ('ws_terms', 'ws_collocations'):
                    corp = self._corp()
                    try: doc = corp.get_struct(corp.get_conf('DOCSTRUCTURE'))
                    except: raise ConcError('DOCSTRUCTURE not set correctly')
                out = corplib.compute_freqfile (self._user,
                      subcmiss.args[0], self.wlattr, self.wlnums, self.get_url())
            return { 'processing': out["progress"] == '100' and '99' or out["progress"],
                     'bgjob_notification': out['notifyme'] and 'checked' or '',
                     'jobid': out["jobid"], 'esttime': out["esttime"],
                     'new_maxitems': self.wlmaxitems}

    wlstruct_attr1 = ''
    wlstruct_attr2 = ''
    wlstruct_attr3 = ''

    def make_wl_query(self, random=0):
        qparts = []
        if self.wlpat: qparts.append('%s="%s"' % (self.wlattr, self.wlpat))
        if not self.include_nonwords:
            qparts.append('%s!="%s"' % (self.wlattr,
                                        self._corp().get_conf('NONWORDRE')))
        if self.wlwords:
            qq = ['%s=="%s"' % (self.wlattr, w.strip()) for w in self.wlwords]
            qparts.append('(' + '|'.join(qq) + ')')
        for w in self.blacklist:
            qparts.append('%s!=="%s"' % (self.wlattr, w.strip()))
        if random: prefix = 'R'
        else: prefix = ''
        self.q = [prefix + 'q[' + '&'.join(qparts) + ']']

    def struct_wordlist (self, random=0):
        if self.wlusesubcorp: usesubcorp = self.usesubcorp = self.wlusesubcorp
        if self.fcrit:
            self.wlwords, self.wlcache = self.get_wl_words()
            self.blacklist, self.blcache = self.get_wl_words(('wlblacklist',
                                                                    'blcache'))
            self.make_wl_query(random)
            return self.freqs (self.fcrit, self.flimit, self.freq_sort, 1)

        if '.' in self.wlattr:
            raise ConcError('Text types are limited to Simple output')
        if self.wlnums != 'frq':
            raise ConcError('Multilevel lists are limited to Word counts frequencies')
        if self.wlattr == 'ws_collocations':
            raise ConcError('Word sketch collocations are not compatible ' +\
                            'with Multilevel wordlist')
        level = 3
        self.wlwords, self.wlcache = self.get_wl_words()
        self.blacklist, self.blcache = self.get_wl_words(('wlblacklist',
                                                                    'blcache'))
        if not self.wlstruct_attr1:
            raise ConcError('No output attribute specified')
        if not self.wlstruct_attr3: level = 2
        if not self.wlstruct_attr2: level = 1
        if not self.wlpat and not self.wlwords:
            raise ConcError('You must specify either a regular expression or a file to get the multilevel wordlist')
        self.make_wl_query(random)
        self.flimit = self.wlminfreq
        return  self.freqml (flimit=self.wlminfreq, freqlevel=level,
                ml1attr=self.wlstruct_attr1, ml2attr=self.wlstruct_attr2,
                ml3attr=self.wlstruct_attr3)

    subcnorm = 'freq'

    def texttypes_with_norms(self, subcorpattrs='', list_all=False, ret_nums=True):
        corp = self._corp()
        if not subcorpattrs:
            subcorpattrs = corp.get_conf ('SUBCORPATTRS') \
                               or corp.get_conf ('FULLREF')
        if not subcorpattrs or subcorpattrs == '#':
            return {'Normslist': [], 'Blocks': []}
        tt = corplib.texttype_values(corp, subcorpattrs, list_all, self.hidenone)
        if not ret_nums: return {'Blocks': tt, 'Normslist': []}
        basestructname = subcorpattrs.split('.')[0]
        struct = corp.get_struct (basestructname)
        normvals = {}
        if self.subcnorm not in ('freq', 'tokens'):
            try:
                nas = struct.get_attr (self.subcnorm).pos2str
            except conclib.manatee.AttrNotFound as e:
                self.error = str(e)
                self.subcnorm = 'freq'
        # TODO: do we need norms here in Bonito4?
        for item in tt:
            for col in item['Line']:
                if 'textboxlength' in col: continue
                if not col['name'].startswith(basestructname + '.'): # new structure
                    basestructname = col['name'].split('.')[0]
                    struct = corp.get_struct (basestructname)
                    normvals = {}
                    if self.subcnorm not in ('freq', 'tokens'):
                        try:
                            nas = struct.get_attr (self.subcnorm).pos2str
                        except conclib.manatee.AttrNotFound as e:
                            self.error = str(e)
                            self.subcnorm = 'freq'
                attr = struct.get_attr(col['name'].split('.')[-1])
                p2i = attr.pos2id
                xcnt_dict = dict([(i, 0) for i in range(attr.id_range())])
                if self.subcnorm == 'freq':
                    for vid in range(attr.id_range()):
                        xcnt_dict[vid] = attr.freq(vid)
                elif self.subcnorm == 'tokens':
                    for sid in range(struct.size()):
                        vid = p2i(sid)
                        xcnt_dict[vid] += struct.end(sid) - struct.beg(sid)
                else:
                    for vid in range(attr.id_range()):
                        xcnt_dict[vid] = attr.norm(vid)
                if not col.get('hierarchical', ''):
                    for val in col['Values']:
                        val['xcnt'] = xcnt_dict.get(attr.str2id(str(val['v'])), 0)
        return {'Blocks': tt, 'Normslist': self.get_normslist(basestructname)}

    def get_normslist(self, structname):
        corp = self._corp()
        normsliststr = corp.get_conf ('DOCNORMS')
        normslist = [{'n':'freq', 'label': 'Document counts'},
                     {'n':'tokens', 'label': 'Tokens'}]
        if normsliststr:
            normslist += [{'n': n, 'label': corp.get_conf (structname + '.'
                                                          + n + '.LABEL') or n}
                          for n in normsliststr.split(',')]
        else:
            try:
                corp.get_attr(structname + ".wordcount")
                normslist.append({'n':'wordcount', 'label': 'Word counts'})
            except:
                pass
        return normslist

    def _texttype_query (self):
        scas = [(a[4:], getattr (self, a))
                for a in dir(self) if a.startswith ('sca_')]
        fscas = [(a[5:], getattr (self, a))
                 for a in dir(self) if a.startswith ('fsca_')]
        structs = {}
        for sa, v in scas: # checkbox input
            if '.' in sa:
                s, a = sa.split('.')
                if isinstance(v, list):
                    query = '(%s)' % ' | '.join (['%s="%s"' % (a,escape(v1))
                                                  for v1 in v])
                else:
                    query = '%s="%s"' % (a,escape(v))
            else: # structure number
                s = sa
                if isinstance(v, list):
                    query = '(%s)' % ' | '.join (['#%s' % x.split('#')[1] for x in v])
                else:
                    query = '#%s' % v.split('#')[1]
            if s not in structs: structs[s] = []
            structs[s].append (query)
        for sa, v in fscas: # free text input (allows REs)
            s, a = sa.split('.')
            query = '%s="%s"' % (a, v)
            if s not in structs: structs[s] = []
            structs[s].append (query)
        return tuple([(sname, ' & '.join(subquery)) for
                sname, subquery in list(structs.items())])
        
    def subcorp (self, subcname='', delete='', create=False, q='', struct='', ttq=''):
        if delete:
            base = os.path.join (self.subcpath[-1], self.corpname,
                                 self.urlencode(subcname))
            base = base
            for e in glob.glob(base + '.*'):
                if os.path.isfile(e) and not e.endswith('.subcdef'):
                    os.unlink(e)
        tt_query = ttq or self._texttype_query()
        if create and not subcname:
            raise ConcError ('No subcorpus name specified!')
        if (not subcname or (not tt_query and delete)
            or (not delete and not create)):
            subcList = self.cm.subcorp_names (self.corpname)
            if subcList and not subcname:
                subcname = subcList[0]['n']
            return {'subcname': subcname,
                    'SubcorpList': self.cm.subcorp_names (self.corpname)}
        if self._anonymous:
            return {'error': 'Anonymous users are not allowed to create subcorpora.'}
        basecorpname = self.corpname.split(':')[0]
        path = os.path.join (self.subcpath[-1], basecorpname)
        if not os.path.isdir (path):
            os.makedirs (path)
        path = os.path.join (path, self.urlencode(subcname)) + '.subc'
        if self._curr_corpus is None:
            self._corp()
        # XXX ignoring more structures
        if q: # AJAX
            conc = self.call_function (conclib.get_conc,
                    (self._curr_corpus, self.samplesize))
            conc.sync()
            s = self._curr_corpus.get_struct(struct)
            cs = conclib.manatee.create_subcorpus (path, conc.RS(), s)
            if cs:
                corplib.save_subcdef(path, subcname,
                                struct or '--NONE--', 'Q:' + '\v'.join(self.q))
                return {'subcorp': subcname, 'message': "Subcorpus saved."}
            else:
                return {'error': "Nothing saved: empty subcorpus!"}
        else:
            if not tt_query:
                raise ConcError('Nothing specified!')
            structname, subquery = tt_query[0]
            if conclib.manatee.create_subcorpus(path, self._curr_corpus,
                    structname, subquery):
                # save subc def file if necessary
                subcdeff = path + "def"
                if not os.path.isfile(subcdeff):
                    subcdeff = corplib.save_subcdef(path,
                            subcname, structname, subquery)
                finalname = '%s:%s' % (basecorpname, subcname)
                sc = self.cm.get_Corpus (finalname)
                return {'subcorp': finalname,
                        'subcorp_deffile': subcdeff,
                        'corpsize': sc.size(),
                        'subcsize': sc.search_size(),
                        'SubcorpList': self.cm.subcorp_names (self.corpname)}
            else:
                raise ConcError('Empty subcorpus!')

    def subcorp_info (self, subcname=''):
        sc = self.cm.get_Corpus (self.corpname, subcname)
        return {'subcorp': subcname,
                'corpsize': sc.size(),
                'subcsize': sc.search_size()}

    def rebuild_subc(self, subcname=''):
        import urllib.request, urllib.parse, urllib.error
        qsubcname = urllib.parse.quote_plus(subcname)
        subcp = os.path.join(self.subcpath[-1], self.corpname,
                qsubcname + '.subcdef')
        subcn, subcs, subcq = corplib.parse_subcdef(qsubcname, subcp)
        self.subcorp(subcname, True)
        if all([subcn, subcs, subcq]):
            subcs = subcs.replace('--NONE--', '')
            if subcq.startswith('Q:'):
                self.q = subcq[2:].split('\v')
                return self.subcorp(subcname, False, True, subcq[2:], subcs, '')
            else:
                return self.subcorp(subcname, False, True, '', '', [(subcs, subcq)])
        else:
            raise RuntimeError('Failed to rebuild subcorpus')

    def corp_info(self, gramrels=0, corpcheck=0, registry=0, subcorpora=0,
            struct_attr_stats=0):
        out = corplib.get_corp_info(self._corp(), registry, gramrels,
                corpcheck, struct_attr_stats)
        compiled = int(out['sizes'].get('tokencount', 0)) > 1
        if compiled and subcorpora:
            tokens = int(out['sizes']['tokencount'])
            words = int(out['sizes'].get('wordcount', tokens))
            wtr = int(words) / float(tokens)
            scs = self.cm.subcorp_names(self.corpname)
            for sc in scs:
                sub = self.cm.get_Corpus(self.corpname, sc['n'])
                subc_size = sub.search_size()
                sc['tokens'] = subc_size
                sc['relsize'] = subc_size / float(tokens) * 100
                sc['words'] = int(subc_size * wtr)
                sc['struct'] = hasattr(sub, 'subcdef') and sub.subcdef[1] or ""
                sc['query'] = hasattr(sub, 'subcdef') and sub.subcdef[2] or ""
            out['subcorpora'] = scs
        else:
            out['subcorpora'] = []
        return out

    def attr_vals(self, avattr='', avpat='', avmaxitems=20, avfrom=0):
        if not avattr:
            return {'error': 'Parameter avattr is required'}
        return corplib.attr_vals(self.abs_corpname(self.corpname), avattr,
                avpat, avmaxitems, avfrom)

    def _storeconc_path(self):
        return os.path.join(self._corp()._conc_dir, self.annotconc)
    
    def storeconc (self, storeconcname=''):
        if self.concordance_query: self.process_concordance_query()
        conc = self.call_function(conclib.get_conc, (self._corp(),))
        conc.sync()
        self.annotconc = storeconcname
        cpath = self._storeconc_path()
        cdir = os.path.dirname(cpath)
        if os.path.exists(cpath + '.conc'):
            return {'stored': storeconcname, 'new': False}
        if not os.path.isdir(cdir):
            os.makedirs(cdir)
        conc.save(cpath + '.conc')
        if self.q[0].startswith('a'):
            self.q[0] = self.q[0].split(',', 1)[1]
        qq = ';'.join(self.q)
        rq = [qi[1:] for qi in self.q if qi.startswith('r')]
        ss = rq and int(rq[0]) or 0
        from skema import Skema
        sk = Skema(self._corp()._conc_dir)
        resp = sk.store_query(concname=storeconcname, queryname=storeconcname,
                query=qq, size=conc.size(), sample=ss, creator=self._user)
        return {'stored': storeconcname, 'resp': resp, 'new': True}

    def delstored(self, annotconc=''):
        cpath = self._storeconc_path()
        from skema import Skema
        sk = Skema(self._corp()._conc_dir)
        do = sk.del_conc(annotconc)
        try:
            os.unlink(cpath + '.conc')
        except OSError:
            pass
        o = {}
        o['message'] = 'Successfully removed'
        o['skema'] = do.get('message', 'Query removed from DB')
        o['removed'] = annotconc
        return o

    def _selectstored (self, annotconc):
        if os.path.exists (self._storeconc_path() + '.conc'):
            return True
        return False

    def storedconcs(self):
        import annotlib
        return annotlib.get_stored_concs(self._corp())

    def _get_annotconc (self):
        import annotlib
        return annotlib.get_stored_conc(self._corp(), self.annotconc)

    def addlngroup(self, label=""):
        import annotlib
        if not label:
            return {'error': 'Missing label'}
        return annotlib.add_new_label(self._corp(), self.annotconc, label)

    def setlngroup(self, toknum='', group=0):
        if not self.annotconc:
            return {'error': 'No concordance selected'}
        conc = self._get_annotconc()
        l = 0
        for tn in set(toknum.strip().split()):
            conc.set_linegroup_at_pos(int(tn), group)
            l += 1
        conc.save (self._storeconc_path() + '.conc', 1)
        lab = conc.labelmap.get (group, group)
        return {'label': lab, 'count': l}

    def setlngroup_example(self, annotconc="", toknum=-1, label=""):
        annot = self._get_annotconc()
        self.q = ["q[#%s]" % str(toknum)]
        conc = self.call_function(conclib.get_conc, (self._corp(),))
        conc.sync()
        conc.set_linegroup_from_conc(annot)
        kl = conclib.manatee.KWICLines(conc.corp(), conc.RS(True), '-1:s',
                '1:s', 'word', 'word', 'g', '#')
        l, k, r = [], [], []
        while kl.nextline():
            l = kl.get_left()
            k = kl.get_kwic()
            r = kl.get_right()
            break
        ls = "".join([l[i] for i in range(0, len(l), 2)]).strip()
        ks = "".join([k[i] for i in range(0, len(k), 2)]).strip()
        rs = "".join([r[i] for i in range(0, len(r), 2)]).strip()
        try:
            from skema import Skema
            sk = Skema(self._corp()._conc_dir)
            return sk.add_example(self.annotconc, label, toknum, ' '.join([ls, ks, rs]))
        except Exception as e:
            return {'error': 'Saving example failed', 'reason': repr(e)}

    def setlngroupglobally (self, group=0):
        if not self.annotconc:
            return 'No concordance selected'
        annot = self._get_annotconc()
        conc = self.call_function (conclib.get_conc, (self._corp(),))
        conc.set_linegroup_from_conc (annot)
        conc.set_linegroup_globally (group)
        annot.set_linegroup_from_conc (conc)
        annot.save (self._storeconc_path() + '.conc', 1)
        lab = annot.labelmap.get (group, group)
        return {'label': lab, 'count': conc.size()}

    def lngroupinfo(self, annotconc=''):
        import annotlib
        return {'labels': annotlib.lngroupinfo(self._corp(), annotconc)}

    minbootscore=0.5
    minbootdiff=0.8
    def bootstrap (self, annotconc='', minbootscore=0.5, minbootdiff=0.8):
        import wsclust
        annot = self._get_annotconc()
        ws = self.call_function (wsclust.WSCluster, ())
        ws.build_pos2coll_map()
        log = ws.bootstrap_conc (annot, minbootscore, minbootdiff)
        annot.save (self._storeconc_path() + '.conc', 1)
        del annot
        self.q = ['s' + annotconc]
        out = self.lngroupinfo (annotconc)
        out['auto_annotated'] = len(log)
        return out

    def fcs(self, operation='explain', version='', recordPacking='xml',
            extraRequestData='', query='', startRecord='', responsePosition='',
            recordSchema='', maximumRecords='', scanClause='', maximumTerms='',
            stylesheet=''):
        "Federated content search API function (www.clarin.eu/fcs)"

        # default values
        self._headers['Content-Type'] = 'application/xml'
        numberOfRecords = 0
        current_version = 1.2
        maximumRecords_ = 250
        maximumTerms_ = 100
        startRecord_ = 1
        responsePosition_ = 1
        # supported parameters for all operations
        sup_pars = ['operation', 'stylesheet', 'version', 'extraRequestData']
        out = {'operation': operation, 'version': current_version,
               'recordPacking': recordPacking, 'result': [],
               'error': False, 'numberOfRecords': numberOfRecords,
               'server_name': self.environ.get('SERVER_NAME', ''),
               'server_port': self.environ.get('SERVER_PORT', '80'),
               'database': self.environ.get('SCRIPT_NAME', '')[1:] + '/fcs',
               'xml_stylesheet': stylesheet,
               }
        fcs_err = None
        try:
            # check version
            if version and current_version < float(version):
                fcs_err = Exception(5, version, 'Unsupported version')

            # check integer parameters
            if maximumRecords != '': 
                try:
                    maximumRecords_ = int(maximumRecords)
                except:
                    fcs_err = Exception(6, 'maximumRecords', 
                                        'Unsupported parameter value')
                if maximumRecords_ < 0:
                    fcs_err = Exception(6, 'maximumRecords', 
                                        'Unsupported parameter value')
                    
            out['maximumRecords'] = maximumRecords_
            if maximumTerms != '':
                try:
                    maximumTerms_ = int(maximumTerms)
                except:
                    fcs_err = Exception(6, 'maximumTerms', 
                                        'Unsupported parameter value')
            out['maximumTerms'] = maximumTerms_
            if startRecord != '':
                try:
                    startRecord_ = int(startRecord)
                except:
                    fcs_err = Exception(6, 'startRecord', 
                                        'Unsupported parameter value')
                if startRecord_ <= 0:
                    fcs_err = Exception(6, 'startRecord',
                                        'Unsupported parameter value')
            out['startRecord'] = startRecord_
            if responsePosition != '':
                try:
                    responsePosition_ = int(responsePosition)
                except:
                    fcs_err = Exception(6, 'responsePosition', 
                                        'Unsupported parameter value')
            out['responsePosition'] = responsePosition_
            
            # set content-type in HTTP header
            if recordPacking == 'string': 
                self._headers['Content-Type'] = 'text/plain'
            elif recordPacking == 'xml':
                self._headers['Content-Type'] = 'application/xml'
            else:
                fcs_err = Exception(71, 'recordPacking', 
                                    'Unsupported record packing')

            # provide info about service
            if operation == 'explain' or not operation:
                sup_pars.append('recordPacking') # other supported parameters
                unsup_pars = list(set(self._url_parameters) - set(sup_pars))
                if unsup_pars:
                    raise Exception(8, unsup_pars[0], 'Unsupported parameter')
                #if extraRequestData:
                #    corpname = extraRequestData
                corp = self._corp()
                out['result'] = corp.get_conf('ATTRLIST').split(',')
                out['numberOfRecords'] = len(out['result'])

            # wordlist for a given attribute
            elif operation == 'scan': 
                # special handling of fcs.resource = root
                if scanClause.startswith('fcs.resource'):
                    out['operation'] = 'fcs.resource'
                    out['resource_info'] = False
                    copts = []
                    if (hasattr(self, 'x-cmd-resource-info') 
                        and getattr(self, 'x-cmd-resource-info') == 'true'):
                        copts = ['DESCRIPTION', 'INFOHREF', 'LANGUAGE']
                        out['resource_info'] = True
                    corps = self.cm.corplist_with_names(copts)
                    out['FCSCorpora'] = corps[:maximumTerms_]
                else:
                    # check supported parameters
                    sup_pars.extend(['scanClause', 'responsePosition',
                                     'maximumTerms'])
                    unsup_pars = list(set(self._url_parameters) - set(sup_pars))
                    if unsup_pars:
                        raise Exception(8, unsup_pars[0], 'Unsupported parameter')
                    #if extraRequestData:
                    #    corpname = extraRequestData
                    out['result'] = conclib.fcs_scan(
                        self.abs_corpname(self.corpname), scanClause,
                        maximumTerms_, responsePosition_)

            # simple concordancer
            elif operation == 'searchRetrieve': 
                # check supported parameters
                sup_pars.extend(['query', 'startRecord', 'maximumRecords', 
                        'recordPacking', 'recordSchema', 'resultSetTTL',
                                 'x-cmd-context'])
                unsup_pars = list(set(self._url_parameters) - set(sup_pars))
                if unsup_pars:
                    raise Exception(8, unsup_pars[0], 'Unsupported parameter')
                if hasattr(self, 'x-cmd-context'):
                    corpname = getattr(self, 'x-cmd-context')
                    if corpname in self.corplist:
                        self._curr_corpus = None
                        self.corpname = corpname
                out['result'] = conclib.fcs_search(self._corp(), query, 
                        maximumRecords_, startRecord_ -1)
                out['numberOfRecords'] = len(out['result'])

            # unsupported operation
            else:
                out['operation'] = 'explain' # show within explain template
                raise Exception(4, 'operation', 'Unsupported operation')
            if fcs_err:
                out['error'] = True
                out['code'], out['details'], out['msg'] = (
                    fcs_err[0], fcs_err[1], fcs_err[2])
            return out

        # catch exception and amend diagnostics in template
        except Exception as e:
            out['error'] = True
            try: # concrete error, catch message from lower levels
                out['code'], out['details'], out['msg'] = e[0], e[1], e[2]
            except: # general error
                out['code'], out['details'] = 1, repr(e)
                out['msg'] = 'General system error'
            return out

    def extract_keywords(self, ref_corpname='', simple_n=1.0,
            attr='lc', stopwords='0', alnum='0', onealpha='1', ref_usesubcorp='',
            minfreq=5, max_keywords=100, lang_code='en', wlpat='.*',
            include_nonwords='0'):
        if not ref_corpname:
            return {'error': "Reference corpus required"}
        blacklist = []
        if int(stopwords): blacklist = [] # TODO: get stopwords from CA
        alnum = bool(int(alnum))
        include_nonwords = bool(int(include_nonwords))
        onealpha = bool(int(onealpha))
        c = self._corp() # both must have the same encoding
        try:
            rc = self.cm.get_Corpus(ref_corpname, ref_usesubcorp)
            result, total, totalfrq = corplib.subc_keywords_onstr(c, rc,
                    wlminfreq=minfreq,
                    wlmaxitems=max_keywords, simple_n=simple_n,
                    blacklist=blacklist, wlattr=attr,
                    onealpha=onealpha, alnum=alnum, wlpat=wlpat,
                    include_nonwords=include_nonwords)
            retlist = []
            for item in result:
                query = 'q[%s="%s"]' % (attr, item[5])
                ref_query = 'q[%s="%s"]' % (attr, item[5])
                retlist.append({
                        'item': item[5],
                        'score': '%.2f' % item[0],
                        'query': query,
                        'ref_query': ref_query,
                        'frq1': item[3],
                        'frq2': item[4],
                        'rel_frq1': item[1],
                        'rel_frq2': item[2]
                })
            if self.corpname.startswith('user/octerms_trial/') and self._anonymous:
                obfuscate_terms(retlist)
            return {'keywords': retlist,
                    'reference_corpus_name': rc.get_conf('NAME'),
                    'reference_corpus_size': rc.size(),
                    'reference_subcorpus_size': rc.search_size(),
                    'subcorpus_size': c.search_size(),
                    'corpus_size': c.size()}
        except corplib.MissingSubCorpFreqFile as e:
            return corplib.compute_freqfile(self._user, e.args[0], attr, 'frq',
                    self.get_url())
        except Exception as e:
            return {'error': repr(e)}

    def extract_terms(self, ref_corpname='', simple_n=1.0, ref_usesubcorp='',
            max_terms=100, alnum='0', onealpha='1', minfreq=1, stopwords='0',
            lang_code='en', wlpat='.*'):
        if not ref_corpname:
            return {'error': "Reference corpus required"}
        alnum = bool(int(alnum))
        onealpha = bool(int(onealpha))
        blacklist = []
        if int(stopwords): blacklist = [] # TODO: get stopwords from CA
        c = self._corp() # both must have the same encoding
        try:
            rc = self.cm.get_Corpus(ref_corpname, ref_usesubcorp)
            retlist = []
            result, total, totalfrq = corplib.ws_keywords(c, rc, wlminfreq=minfreq,
                    simple_n=simple_n, wlattr='ws_terms', wlpat=wlpat,
                    alnum=alnum, onealpha=onealpha, blacklist=blacklist,
                    wlmaxitems=max_terms*10) # TODO: store only mw terms
            for item in result[:max_terms]:
                sitem = item[5].split('\t')
                query = 't' + sitem[3]
                ref_query = 't' + sitem[4]
                retlist.append({
                        'item': sitem[2].replace('_', ' '),
                        'score': '%.2f' % item[0],
                        'query': query,
                        'ref_query': ref_query,
                        'frq1': item[3],
                        'frq2': item[4],
                        'rel_frq1': item[1],
                        'rel_frq2': item[2],
                        'seek': sitem[3],
                        'ref_seek': sitem[4]
                })
            if self.corpname.startswith('user/octerms_trial/') and self._anonymous:
                obfuscate_terms(retlist)
            return {'terms': retlist,
                    'reference_corpus_name': rc.get_conf('NAME'),
                    'reference_corpus_size': rc.size(),
                    'reference_subcorpus_size': rc.search_size(),
                    'corpus_size': c.size(),
                    'subcorpus_size': c.search_size()}
        except corplib.MissingSubCorpFreqFile as e:
            return corplib.compute_freqfile(self._user, e.args[0], 'ws_terms',
                    'frq', self.get_url())
        except Exception as e:
            return {'error': repr(e)}

    def translate_words(self, l2_corpname='', data=''):
        return conclib.translate_words(self._corp(),
                self.abs_corpname(l2_corpname), data)

    def translate_kwic(self, bim_corpname='', data=''):
        result = conclib.translate_kwic(self._corp(),
                                        self.abs_corpname(bim_corpname), data)
        return {'toknum2words': result[1], 'dict': result[0], 'corpus': bim_corpname}

    def biterms(self, corpname='', l2_corpname=''):
        # always limit to 1000 items, biterms are pre-sorted
        limit = 1000
        biterms = {}
        if '/' in corpname:
            l2_corpname = corpname.rsplit('/', 1)[0] + '/' + l2_corpname
        biterms['aligned_corpname'] = l2_corpname
        try:
            bitgen = list(corplib.get_biterms(self._corp(),
                    self.cm.get_Corpus(l2_corpname), limit))
        except Exception as e:
            biterms['Biterms'] = []
            biterms['BitermsDict'] = {}
            biterms['error'] = repr(e)
            biterms['message'] = "An error occurred when extracting "\
                    "bilingual terminology"
            return biterms
        l = []
        d = {}
        for b in bitgen:
            if not b[0] in l:
                l.append(b[0])
            d.setdefault(b[0], []).append(b)
        if self.corpname.startswith('user/octerms_trial/') and self._anonymous:
            obfuscate_biterms(l, d)
        biterms['Biterms'] = l
        biterms['BitermsDict'] = d
        return biterms

    def trends(self, structattr='', trends_attr='word', usesubcorp='',
            trends_method='mkts_all', trends_re='', trends_sort_by='t',
            filter_capitalized=1, trends_maxp=0.1, filter_nonwords=1,
            trends_minfreq=5, trends_max_items=1000, filter_by_trend=0,
            trends_order='asc'):
        subcpath = getattr(self._corp(), 'spath', '')
        result, samples = corplib.get_trends(self._corp(), structattr,
                trends_attr, subcpath, trends_re, trends_sort_by,
                filter_nonwords, filter_capitalized, trends_method, trends_maxp,
                trends_minfreq, trends_max_items, filter_by_trend, trends_order)
        if result is None: # trends not computed
            missing_norms = False
            missing_freqs = False
            if usesubcorp:
                missing_norms = not os.path.exists(subcpath[:-5] + '.' +\
                        structattr + '.norm')
                missing_freqs = not os.path.exists(subcpath[:-5] + '.' +\
                        trends_attr + '.frq')
            else:
                missing_norms = not os.path.exists(os.path.join(self._corp().get_conf('PATH'),\
                        structattr + '.norm'))
                missing_norms = not os.path.exists(os.path.join(self._corp().get_conf('PATH'),\
                        structattr + '.norm'))
            if missing_freqs:
                return corplib.compute_freqfile(self._user, self._corp(),
                        trends_attr, 'frq', self.get_url())
            if missing_norms:
                return corplib.compute_norms(self._corp(), self._user,\
                        structattr.split('.')[0], subcpath, self.get_url())
            return corplib.compute_trends(self._corp(), self._user,
                    structattr, trends_attr, trends_method, subcpath, self.get_url())
        if result is False: # computing
            return {'processing': 1, 'result': []}
        if not result: # empty result
            return {'result': [], 'message': "No data for these parameters"}
        attr_label = self._corp().get_conf(trends_attr + '.LABEL') or trends_attr
        return {'result': result, 'samples': samples, 'structattr': structattr,
                'trends_attr': trends_attr, 'usesubcorp': usesubcorp,
                'trends_attr_label': attr_label, 'trends_maxp': trends_maxp,
                'trends_method': trends_method, 'trends_sort_by':trends_sort_by}

    def poswordlist(self, wlpat='', wltype='simple', ref_corpname='',
            ref_usesubcorp='', wlpage=1, lpos='', wlicase='0'):
        attrlist = self._corp().get_conf('ATTRLIST').split(',')
        if 'lempos' not in attrlist:
            return {'error': 'PoS wordlist not available for this corpus'}
        if 'lempos_lc' in attrlist and wlicase == '1':
            self.wlattr = 'lempos_lc'
        else:
            self.wlattr = 'lempos'
        self.wlpat = (wlpat or '.*') + lpos
        poswl = self.wordlist(self.wlpat, wltype, self.corpname, self.usesubcorp,
                ref_corpname, ref_usesubcorp, wlpage=wlpage)
        for i in poswl.get('Items', []):
            i['str'] = i['str'][:-2]
        return poswl

    def wsposlist(self):
        return {'wsposlist': [(x['n'], x['v'])\
                for x in corplib.get_wsposlist(self._corp())]}

    def wikisearch(self, query='', langid='en'):
        import urllib.request, urllib.parse, urllib.error, urllib.request, urllib.error, urllib.parse, json
        url = 'http://%s.wikipedia.org/w/api.php?action=query&list=search&'\
                'format=json&srsearch=%s'#&srprop=titlesnippet'
        h = {'User-Agent': 'Sketch Engine Terminology Extraction/1.0 '\
                '(https://app.sketchengine.eu; support@sketchengine.co.uk)'}
        req = urllib.request.Request(url % (langid, urllib.parse.quote_plus(query)),
                headers=h)
        f = urllib.request.urlopen(req)
        d = f.read()
        f.close()
        return json.loads(d)
