#coding=utf-8

## Copyright (c) 2019 Vít Baisa

import sys
import json as jsonlib

LTR = '\xe2\x80\x8e'
RTL = '\xe2\x80\x8f'

def s(o):
    return o.get('str', o.get('strc', '')) + o.get('attr', '')

def n(s):
    # TODO: encode into corpus encoding (not UTF-8)
    return s.replace('&', '&amp;')\
            .replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')\
            .replace('  ', ' ')

def c(s):
    # TODO: encode into corpus encoding (not UTF-8)
    return s.replace('"', '""').replace("'", "''")\
            .replace('  ', ' ')

def t(s):
    # TODO: encode into corpus encoding (not UTF-8)
    return s.replace('  ', ' ').replace('\t', '\v')

def txt(methodname, data, outf, nl, params):
    outf.write('corpus: %s' % t(params.get('corpname', '')) + nl)
    outf.write('subcorpus: %s' % t(params.get('usesubcorp', '-')) + nl)
    if data.get('concsize', False):
        outf.write('size: %d' % int(data['concsize']) + nl)
    if (data.get('Desc', '')):
        outf.write('query: %s' % ', '.join([t(d.get('op', 'operation')) + ':'\
                + t(d.get('arg', 'argument')) for d in data['Desc']]) + nl)
    if 'error' in data:
        outf.write('An error occurred, check results in UI or in JSON format.' + nl)
        return
    if methodname in ['view', 'concordance', 'first']:
        def printl(l):
            if l.get('linegroup', '_') != "_":
                outf.write('#' + t(l['linegroup']) + ' ')
            if l.get('Refs', []):
                outf.write('%s| ' % t(','.join(l['Refs'])))
            outf.write(' '.join([t(s(x)) for x in l['Left']]) + ' ')
            if params.get('sepkwic', False):
                outf.write('<<<')
            outf.write(' '.join([t(s(x)) for x in l['Kwic']]))
            if params.get('sepkwic', False):
                outf.write('>>>')
            outf.write(' ')
            outf.write(' '.join([t(s(x)) for x in l['Right']]) + ' ')
        for line in data['Lines']:
            printl(line)
            if line.get('Align', []):
                outf.write('\t')
                for j, al in enumerate(line['Align']):
                    printl(al)
            outf.write(nl)
    elif methodname.startswith('freq') or methodname == 'struct_wordlist':
        for block in data['Blocks']:
            if block.get('blockname', ''):
                outf.write('%s:' % block['blockname'] + nl)
                outf.write('-------' + nl)
            for item in block['Items']:
                for w in item['Word']:
                    outf.write('%s, ' % t(w['n']))
                outf.write('%d' % item['freq'])
                if 'rel' in item:
                    outf.write(', %.5f' % item['rel'])
                if 'fpm' in item:
                    outf.write(', %.5f' % item['fpm'])
                if params.get('examples', ''):
                    outf.write(', %s' % t(item.get('example', '')))
                outf.write(nl)
    elif methodname.startswith('coll'):
        for item in data['Items']:
            outf.write('%s	%d %d ' % (c(item['str']),
                    int(item['freq']), int(item['coll_freq'])))
            outf.write(' '.join(['%.5f' % float(x['s'])\
                    for x in item['Stats']]))
            outf.write(nl)
    else:
        outf.write("Method %s doesn't support export to TXT." % methodname + nl)

def tmx(methodname, data, outf, nl, params):
    outf.write("<?xml version='1.0' encoding='UTF-8' ?>" + nl)
    outf.write('<tmx version="1.0">' + nl)
    ouff.write('  <header creationtool="Sketch Engine" segtype="sentence"')
    outf.write(' datatype="plaintext" srclang="??">' + nl)
    outf.write('  </header>' + nl)
    outf.write('  <body>' + nl)
    if 'error' in data:
        outf.write('    <error>An error occurred, check results in UI or in JSON format.</error>' + nl)
        outf.write('</body></tmx>' + nl)
        return
    if methodname in ['view', 'concordance', 'first']:
        def printl(l):
            outf.write('      <tuv lang="??">' + nl)
            outf.write('        <seg>')
            outf.write(' '.join([n(s(x)) for x in l['Left']]) + ' ')
            outf.write(' '.join([n(s(x)) for x in l['Kwic']]))
            outf.write(' '.join([n(s(x)) for x in l['Right']]) + ' ')
            outf.write('        </seg>' + nl)
            outf.write('      </tuv>' + nl)
        for line in data['Lines']:
            outf.write('    <tu>' + nl)
            printl(line)
            if line.get('Align', []):
                outf.write('')
                for j, al in enumerate(line['Align']):
                    printl(al)
            outf.write('    </tu>' + nl)
    else:
        outf.write('    <tu><tuv lang="en">')
        outf.write('<seg>Method %s does not support export to TMX.' % methodname)
        outf.write('</seg></tuv></tu>' + nl)
    outf.write('  </body>' + nl)
    outf.write('</tmx>' + nl)

def csv(methodname, data, outf, nl, params):
    # TODO: start with BOM if saving for Windows?
    outf.write('"corpus","%s"' % c(params.get('corpname', '')) + nl)
    outf.write('"subcorpus","%s"' % c(params.get('usesubcorp', '-')) + nl)
    if data.get('concsize', False):
        outf.write('"concordance size","%d"' % int(data['concsize']) + nl)
    if (data.get('Desc', '')):
        outf.write('"query","%s"' % ';'.join(
                [c(d.get('op', 'operation')) + ':' + c(d.get('arg', 'argument'))\
                for d in data['Desc']]) + nl)
    if 'error' in data:
        outf.write('"An error occurred, check results in UI or in JSON format."' + nl)
        return
    if methodname == 'biterms':
        outf.write('"Source","Target","Score","Co-hits","Source hits",'
                '"Target hits"' + nl)
        for item in data['Biterms']:
            b = data['BitermsDict'][item][0]
            outf.write('"%s","%s",%.3f,%d,%d,%d' % (c(b[0]), c(b[1]),\
                    float(b[7]), int(b[2]), int(b[3]), int(b[4])))
            outf.write(nl)
    elif methodname == 'wsketch':
        for gr in data['Gramrels']:
            nicename = c(gr['name'].replace('"%w"', 'X').replace('%w', 'X'))
            outf.write('"%s",%d,%.3f' % (nicename, gr['count'] or 0,
                    float(gr['score'] or '0.0') or 0.0))
            outf.write(nl)
            for w in gr['Words']:
                outf.write(',"%s",%d,%.3f' %\
                        (c(w.get('word', w.get('name', ''))),
                        int(w['count']), float(w['score'])))
                outf.write(nl)
    elif methodname.startswith('coll'):
        for item in data['Items']:
            outf.write('"%s",%d,%d,' % (c(item['str']),
                    int(item['freq']), int(item['coll_freq'])))
            outf.write(','.join(['%.5f' % float(x['s'])\
                    for x in item['Stats']]))
            outf.write(nl)
    elif methodname in ['view', 'first', 'concordance']:
        def printl(l):
            if l.get('linegroup', '_') != "_":
                outf.write('"' + t(l['linegroup']) + '",')
            if data.get('righttoleft', False):
                l['Left'], l['Right'] = l['Right'], l['Left']
            outf.write('"%s",' % c(','.join(line.get('Refs', []))))
            if data.get('righttoleft', False):
                outf.write(RTL)
            outf.write('"' + ' '.join([c(s(x)) for x in l['Left']]) + '","')
            outf.write(' '.join([c(s(x)) for x in l['Kwic']]) + '","')
            outf.write(' '.join([c(s(x)) for x in l['Right']]) + '"')
            if data.get('righttoleft', False):
                outf.write(LTR)
        for line in data['Lines']:
            printl(line)
            if line.get('Align', []):
                outf.write(',')
                for j, al in enumerate(line['Align']):
                    printl(al)
            outf.write(nl)
    elif methodname.startswith('freq') or methodname == 'struct_wordlist':
        for block in data['Blocks']:
            if block.get('blockname', ''):
                outf.write('BLOCK %s:' % block['blockname'] + nl)
            for item in block['Items']:
                for w in item['Word']:
                    outf.write('"%s",' % c(w['n']))
                outf.write('%d' % item['freq'])
                if 'rel' in item:
                    outf.write(',%.5f' % item['rel'])
                if 'fpm' in item:
                    outf.write(', %.5f' % item['fpm'])
                if params.get('examples', ''):
                    outf.write(',"%s"' % c(item.get('example', '')))
                outf.write(nl)
    elif methodname == 'thes':
        outf.write('"%s",%d,"%s"' %\
                (c(params.get('lemma', '')), params.get('freq', 0),
                c(params.get('lpos', ''))) + nl)
        for w in data['Words']:
            outf.write('"%s",%.3f,%d' %\
                    (c(w['word']), w['score'], w['freq']) + nl)
            for cl in w.get('Clust', []):
                outf.write(',"%s",%.3f,%d' %\
                        (c(cl['word']), cl['score'], cl['freq']) + nl)
    elif methodname == 'wipo':
        # TODO: add automatic test?
        outf.write("Extracted terms" + nl)
        outf.write("Corpus" + ': %s' % params.get('corp_full_name', '') + nl)
        outf.write(nl)
        outf.write(','.join(["Term", "Frequency", "Freq/mill", "Score"]) + nl)
        for k in data['Keyterms']:
            outf.write(','.join([c(k['item']), str(k['frq1']), str(k['relfrq1']), str(k['score'])]) + nl)
    elif methodname in ['wordlist', 'poswordlist']:
        # TODO: these normalizations shouldn't be needed in the future
        usengrams = params.get('usengrams', "0")
        usengrams = type(usengrams) == type([]) and usengrams[-1] or usengrams
        nest_ngrams = params.get('nest_ngrams', "0")
        nest_ngrams = type(nest_ngrams) == type([]) and nest_ngrams[-1] or\
                nest_ngrams
        def nested(ngram, level=0):
            if level: outf.write('"')
            for i in range(level):
                outf.write('... ')
            if level:
                outf.write('%s",%d' % (c(ngram['str']), ngram['freq']) + nl)
            for ch in ngram.get('children', []):
                nested(ch, level+1)
        if data.get('Items', []):
            outf.write('"%s","Freq"' % params.get('wlattr', '') + nl)
            for i in data['Items']:
                if params.get('wlattr', '') == 'ws_terms':
                    outf.write('"%s",%d,' % (c(i['w2'].replace('_', ' ')),\
                            i['freq']) + nl)
                elif params.get('wlattr', '') == 'ws_collocations':
                    outf.write('"%s",%d,%.3f' %\
                            (c(i['str']), i['freq'], float(i['score'] or '0.0'))\
                            + nl)
                else:
                    if usengrams == "1" and nest_ngrams == "1" and\
                            i.get('nested', ''):
                        continue
                    outf.write('"%s",%d' % (c(i['str']), i['freq']) + nl)
                    if usengrams == "1" and nest_ngrams == "1" and\
                            i.get('children', []):
                        nested(i)
        elif data.get('Keywords', []):
            outf.write('"%s","Freq","Freq/mill","Freq_ref","Freq_ref/mill",'
                    '"Score"' + nl)
            for k in data['Keywords']:
                if params.get('wlattr', "0") == 'ws_terms':
                    w = c(k['str']).split('\t')[2].replace('_', ' ')
                else:
                    w = c(k['str'])
                outf.write('"%s",%d,%.3f,%d,%.3f,%.3f' %\
                        (w, k['freq'], k['rel'], k.get('freq_ref', 0),
                        k.get('rel_ref', 0.0), k.get('score', 0.0)) + nl)
    elif methodname == 'wsdiff':
        outf.write('"%s","%s"' % (c(data['lbl1']), c(data['lbl2'])) + nl)
        for gr in data['content']['common']:
            h = gr['table']['Header']
            outf.write('"RELATION %s",%s,%s,%s,%s' % (c(h[0]),
                    h[1], h[2], h[3], h[4]) + nl)
            for r in gr['table']['Rows']:
                outf.write('"%s",%s,%s,%s,%s,%s' % (c(r['word']),
                        r['cnt1'], r['cnt2'], r['rnk1str'], r['rnk2str'],
                        r['class']) + nl)
    elif methodname in ['extract_terms', 'extract_keywords']:
        outf.write('"Item","Score","Freq","Ref_freq","Rel_freq","Rel_ref_freq"' + nl)
        for item in data.get('terms', data.get('keywords', [])):
            outf.write('"%s",%.3f,%d,%d,%.3f,%.3f' % (c(item['item']),
                    float(item['score']), int(item['frq1']), int(item['frq2']),
                    float(item['rel_frq1']), float(item['rel_frq2'])))
            outf.write(nl)
    else:
        outf.write('Method %s does not support export to CSV.' %\
                methodname + nl)

def xls(methodname, data, outf, nl, params):
    outf.write('<html>' + nl)
    outf.write('  <head>' + nl)
    outf.write('    <title>%s</title>' % methodname + nl)
    outf.write('    <meta http-equiv="Content-Type" '
            'content="text/html;charset=UTF-8">' + nl)
    outf.write('  </head>' + nl)
    outf.write('  <body>' + nl)
    outf.write('    <table>' + nl)
    if 'error' in data:
        outf.write('    <tbody><tr><th>An error occurred, check results in UI or in JSON format.</th></tr></tbody>' + nl)
        outf.write('</table></body></html>' + nl)
        return
    if methodname == 'biterms':
        outf.write('      <thead><tr><th>Source</th><th>Target</th>'
                '<th>Score</th><th>Co-freq</th><th>L1 Freq</th>'
                '<th>L2 Freq</th></tr></thead>' + nl)
        outf.write('      <tbody>' + nl)
        for item in data['Biterms']:
            b = data['BitermsDict'][item][0]
            outf.write('        <tr>' + nl)
            outf.write('          <td>%s</td>' % n(b[0]) + nl)
            outf.write('          <td>%s</td>' % n(b[1]) + nl)
            outf.write('          <td>%.3f</td>' % float(b[7]) + nl)
            outf.write('          <td>%d</td>' % int(b[2]) + nl)
            outf.write('          <td>%d</td>' % int(b[3]) + nl)
            outf.write('          <td>%d</td>' % int(b[4]) + nl)
            outf.write('        </tr>' + nl)
        outf.write('      </tbody>' + nl)
    elif methodname in ['extract_terms', 'extract_keywords']:
        outf.write('      <thead><tr><th>Term</th><th>Score</th><th>Freq</th>'
                '<th>Ref freq</th><th>Rel freq</th><th>Rel ref freq</th>'
                '</tr></thead>' + nl)
        outf.write('      <tbody>' + nl)
        for item in data.get('terms', data.get('keywords', [])):
            outf.write('      <tr>' + nl)
            outf.write('        <td>%s</td>' % n(item['item']) + nl)
            outf.write('        <td>%.3f</td>' % float(item['score']) + nl)
            outf.write('        <td>%d</td>' % int(item['frq1']) + nl)
            outf.write('        <td>%d</td>' % int(item['frq2']) + nl)
            outf.write('        <td>%.3f</td>' % float(item['rel_frq1']) + nl)
            outf.write('        <td>%.3f</td>' % float(item['rel_frq2']) + nl)
            outf.write('      </tr>' + nl)
        outf.write('      </tbody>' + nl)
    elif methodname == 'wsketch':
        outf.write('      <thead><tr><th>Grammar relation</th>'
                '<th>Collocate</th><th>Freq</th>'
                '<th>Score</th></tr></thead>' + nl)
        outf.write('      <tbody>' + nl)
        for gr in data['Gramrels']:
            nicename = n(gr['name'].replace('"%w"', 'X').replace('%w', 'X'))
            outf.write('      <tr>' + nl)
            outf.write('        <td>%s</td>' % nicename + nl)
            outf.write('        <td>%d</td>' % (gr['count'] or 0) + nl)
            outf.write('        <td>%.3f</td>' %\
                    (float(gr['score'] or '0.0') or 0.0) + nl)
            outf.write('        <td></td>' + nl)
            outf.write('      </tr>' + nl)
            outf.write('      <tr>' + nl)
            for w in gr['Words']:
                outf.write('        <td></td>' + nl)
                outf.write('        <td>%s</td>' % n(w.get('word', w.get('name', ''))) + nl)
                outf.write('        <td>%d</td>' % w['count'] + nl)
                outf.write('        <td>%.3f</td>' % w['score'] + nl)
                outf.write('      </tr>' + nl)
        outf.write('</tbody>' + nl)
    elif methodname.startswith('coll'):
        mldict = {x['s']:x['n'] for x in data['Head'] if 's' in x}
        outf.write('      <thead><tr><th>Collocate</th>'
                '<th>Collocate</th><th>Freq</th>'
                '<th>Coll. freq.</th>')
        for x in data['Items'][0]['Stats']:
            outf.write('<th>%s</th>' % mldict.get(x['n'], ''))
        outf.write('</tr></thead>' + nl)
        outf.write('      <tbody>' + nl)
        for item in data['Items']:
            outf.write('      <tr>' + nl)
            outf.write('        <td>%s</td>' % n(item['str']) + nl)
            outf.write('        <td>%d</td>' % int(item['freq']) + nl)
            outf.write('        <td>%d</td>' % int(item['coll_freq']) + nl)
            for x in item['Stats']:
                outf.write('        <td>%.5f</td>' % float(x['s']) + nl)
            outf.write('      </tr>' + nl)
        outf.write('</tbody>' + nl)
    elif methodname in ['view', 'first', 'concordance']:
        outf.write('      <thead><tr><th>Reference</th>'
                '<th>Left</th><th>Kwic</th>'
                '<th>Right</th></tr></thead>' + nl)
        def printl(l):
            if l.get('linegroup', '_') != "_":
                outf.write('    <td>%s</td>' % n(l['linegroup']) + nl)
            outf.write('        <td>%s</td>' % n(','.join(l.get('Refs', []))) + nl)
            left = ' '.join([n(s(x)) for x in l['Left']])
            kwic = ' '.join([n(s(x)) for x in l['Kwic']])
            right = ' '.join([n(s(x)) for x in l['Right']])
            outf.write('        <td>%s</td>' % left + nl)
            outf.write('        <td>%s</td>' % kwic + nl)
            outf.write('        <td>%s</td>' % right + nl)
        for i, line in enumerate(data['Lines']):
            outf.write('      <tr>' + nl)
            if line.get('Align', []):
                printl(line)
                for j, al in enumerate(line['Align']):
                    printl(al)
            else:
                printl(line)
            outf.write('      </tr>' + nl)
        outf.write('    </tbody>' + nl)
    elif methodname.startswith('freq') or methodname == 'struct_wordlist':
        outf.write('    <tbody>' + nl)
        for block in data['Blocks']:
            attrd = {x['s']:n(x['n']) for x in block['Head'] if 'id' in x}
            if block.get('blockname', ''):
                outf.write('<tr><th>%s</th></tr>' % n(block['blockname']) + nl)
            outf.write('    <tr>' + nl)
            for i, w in enumerate(block['Items'][0]['Word']):
                outf.write('      <th>%s</th>' % attrd.get(i, 'str') + nl)
            outf.write('      </tr>' + nl)
            for item in block['Items']:
                outf.write('      <tr>' + nl)
                for i, w in enumerate(item['Word']):
                    outf.write('        <td>%s</td>' % n(w['n']) + nl)
                outf.write('        <td>%d</td>' % int(item['freq']) + nl)
                if 'rel' in item:
                    outf.write('      <td>%.5f</td>' % float(item['rel']) + nl)
                if 'fpm' in item:
                    outf.write('      <td>%.5f</td>' % float(item['fpm']) + nl)
                if params.get('examples', ''):
                    outf.write('      <td>%s</td>' %\
                            n(item.get('example', '')) + nl)
                outf.write('      </tr>' + nl)
        outf.write('  </tbody>' + nl)
    elif methodname == 'thes':
        outf.write('      <thead><tr><th>Lemma</th>'
                '<th>Score</th><th>Freq</th></tr></thead>' + nl)
        outf.write('  <tbody>' + nl)
        for w in data['Words']:
            outf.write('      <tr>' + nl)
            if 'Clust' in w:
                outf.write('      <td>%s</td>' % n(w['word']) + nl)
            outf.write('        <td>%s</td>' % n(w['word']) + nl)
            outf.write('        <td>%.3f</td>' % w['score'] + nl)
            outf.write('        <td>%d</td>' % int(w['freq']) + nl)
            for cl in w.get('Clust', []):
                outf.write('        <td></td>' + nl)
                outf.write('        <td>%s</td>' % n(cl['word']) + nl)
                outf.write('        <td>%.3f</td>' % cl['score'] + nl)
                outf.write('        <td>%d</td>' % int(cl['freq']) + nl)
            outf.write('      </tr>' + nl)
        outf.write('  </tbody>' + nl)
    elif methodname in ['wordlist', 'poswordlist']:
        # TODO: these normalizations shouldn't be needed in the future
        usengrams = params.get('usengrams', "0")
        usengrams = type(usengrams) == type([]) and usengrams[-1] or usengrams
        nest_ngrams = params.get('nest_ngrams', "0")
        nest_ngrams = type(nest_ngrams) == type([]) and nest_ngrams[-1] or\
                nest_ngrams
        def nested(ngram, first=True):
            if not ngram.get('nested', '') or not first:
                outf.write('      <tr>' + nl)
                outf.write('        <td></td>' + nl)
                outf.write('        <td>%s</td>' % n(ngram['str']) + nl)
                outf.write('        <td>%d</td>' % int(ngram['freq']) + nl)
                for ch in ngram.get('children', []):
                    nested(ch, False)
                outf.write('      </tr>' + nl)
        if data.get('Items', []):
            outf.write('    <thead><tr><th>Item</th><th>Freq</th>')
            if params.get('wlattr', '') == 'ws_collocations':
                outf.write('<th>Score</th>')
            outf.write('</tr></thead>' + nl)
            for i in data['Items']:
                if usengrams == "1" and nest_ngrams == "1":
                    nested(i)
                outf.write('        <tr>' + nl)
                if params.get('wlattr', '') == 'ws_terms':
                    outf.write('      <td>%s</td>' %\
                            n(i['w2']).replace('_', ' ') + nl)
                else:
                    outf.write('      <td>%s</td>' % n(i['str']) + nl)
                outf.write('        <td>%d</td>' % int(i['freq']) + nl)
                if params.get('wlattr', '') == 'ws_collocations':
                    outf.write('        <td>%.3f</td>' % float(i['score']) + nl)
                outf.write('        </tr>' + nl)
            outf.write('    </tbody>' + nl)
        if data.get('Keywords', []):
            outf.write('    <thead><tr><th>Keyword</th><th>Freq</th>'
                    '<th>Rel.</th><th>Freq. ref.</th>'
                    '<th>Rel. ref.</th><th>Score</th></tr></thead>' + nl)
            for k in data['Keywords']:
                if params.get('wlattr', '') == 'ws_terms':
                    w = n(k['str'].split('\t')[2]).replace('_', ' ')
                else:
                    w = n(k['str'])
                outf.write('      <tr>' + nl)
                outf.write('        <td>%s</td>' % w + nl)
                outf.write('        <td>%d</td>' % k['freq'] + nl)
                outf.write('        <td>%.3f</td>' % k['rel'] + nl)
                outf.write('        <td>%.3f</td>' % k.get('freq_ref', .0) + nl)
                outf.write('        <td>%.3f</td>' % k.get('rel_ref', .0) + nl)
                outf.write('        <td>%.3f</td>' % k.get('score', .0) + nl)
                outf.write('      </tr>' + nl)
            outf.write('      </tbody>' + nl)
    elif methodname == 'wsdiff':
        outf.write('    <thead><tr><th></th><th>Freq</th>'
                '<th>Rel.</th><th>Freq. ref.</th>'
                '<th>Rel. ref.</th><th>Score</th>'
                '<th>Color</th></tr></thead>' + nl)
        for gr in data['content']['common']:
            h = gr['table']['Header']
            outf.write('        <tr>' + nl)
            outf.write('          <td>%s</td>' % n(h[0]) + nl)
            outf.write('          <td>%d</td>' % int(h[1]) + nl)
            outf.write('          <td>%d</td>' % int(h[2]) + nl)
            outf.write('          <td>%.3f</td>' % float(h[3]) + nl)
            outf.write('          <td>%.3f</td>' % float(h[4]) + nl)
            outf.write('        </tr>' + nl)
            for r in gr['table']['Rows']:
                outf.write('          <tr>' + nl)
                outf.write('            <td>%s</td>' % n(r['word']) + nl)
                outf.write('            <td>%d</td>' % int(r['cnt1']) + nl)
                outf.write('            <td>%d</td>' % int(r['cnt2']) + nl)
                outf.write('            <td>%s</td>' % r['rnk1str'] + nl)
                outf.write('            <td>%s</td>' % r['rnk2str'] + nl)
                outf.write('            <td>%s</td>' % r['class'] + nl)
            outf.write('        </tr>' + nl)
        outf.write('      </tbody>' + nl)
    else:
        outf.write('<thead><tr><th>Method %s does not support export to XLS.'
                '</th></tr></thead>' % methodname + nl)
    outf.write('</table></body></html>' + nl)

def tsv(methodname, data, outf, nl, params):
    outf.write('corpus\t%s' % t(params.get('corpname', '')) + nl)
    outf.write('subcorpus\t%s' % t(params.get('usesubcorp', '-')) + nl)
    if data.get('concsize', False):
        outf.write('concordance size\t%d' % int(data['concsize']) + nl)
    if (data.get('Desc', '')):
        outf.write('query\t%s' % ';'.join(
                [t(d.get('op', 'operation')) + ':' + t(d.get('arg', 'argument'))\
                for d in data['Desc']]) + nl)
    if 'error' in data:
        outf.write('An error occurred, check results in UI or in JSON format.' + nl)
        return
    if methodname == 'biterms':
        outf.write('Source\tTarget\tScore\tCo-hits\tSource hits\tTarget hits'+nl)
        for item in data['Biterms']:
            b = data['BitermsDict'][item][0]
            outf.write('%s\t%s\t%.3f\t%d\t%d\t%d' % (t(b[0]), t(b[1]),\
                    float(b[7]), int(b[2]), int(b[3]), int(b[4])))
            outf.write(nl)
    elif methodname == 'wsketch':
        for gr in data['Gramrels']:
            nicename = t(gr['name'].replace('"%w"', 'X').replace('%w', 'X'))
            outf.write('%s\t%d\t%.3f' % (nicename, gr['count'] or 0,
                    float(gr['score'] or '0.0') or 0.0))
            outf.write(nl)
            for w in gr['Words']:
                outf.write('\t%s\t%d\t%.3f' %\
                        (t(w.get('word', w.get('name', ''))),
                        int(w['count']), float(w['score'])))
                outf.write(nl)
    elif methodname.startswith('coll'):
        for item in data['Items']:
            outf.write('%s\t%d\t%d,' % (t(item['str']),
                    int(item['freq']), int(item['coll_freq'])))
            outf.write('\t'.join(['%.5f' % float(x['s'])\
                    for x in item['Stats']]))
            outf.write(nl)
    elif methodname in ['view', 'first', 'concordance']:
        def printl(l):
            if l.get('linegroup', '_') != "_":
                outf.write('#' + l['linegroup'] + '\t')
            if data.get('righttoleft', False):
                l['Left'], l['Right'] = l['Right'], l['Left']
            outf.write('%s\t' % t(','.join(line.get('Refs', []))))
            if data.get('righttoleft', False):
                outf.write(RTL)
            outf.write(' '.join([t(s(x)) for x in l['Left']]) + '\t')
            outf.write(' '.join([t(s(x)) for x in l['Kwic']]) + '\t')
            outf.write(' '.join([t(s(x)) for x in l['Right']]))
            if data.get('righttoleft', False):
                outf.write(LTR)
        for line in data['Lines']:
            printl(line)
            if line.get('Align', []):
                outf.write('\t')
                for j, al in enumerate(line['Align']):
                    printl(al)
            outf.write(nl)
    elif methodname.startswith('freq') or methodname == 'struct_wordlist':
        for block in data['Blocks']:
            if block.get('blockname', ''):
                outf.write('BLOCK %s:' % block['blockname'] + nl)
            for item in block['Items']:
                for w in item['Word']:
                    outf.write('%s\t' % t(w['n']))
                outf.write('%d' % item['freq'])
                if 'rel' in item:
                    outf.write('\t%.5f' % item['rel'])
                if 'fpm' in item:
                    outf.write('\t%.5f' % item['fpm'])
                if params.get('examples', ''):
                    outf.write('\t%s' % t(item.get('example', '')))
                outf.write(nl)
    elif methodname == 'thes':
        outf.write('%s\t%d\t%s' %\
                (t(params.get('lemma', '')), params.get('freq', 0),
                t(params.get('lpos', ''))) + nl)
        for w in data['Words']:
            outf.write('%s\t%.3f\t%d' %\
                    (t(w['word']), w['score'], w['freq']) + nl)
            for cl in w.get('Clust', []):
                outf.write('\t%s\t%.3f\t%d' %\
                        (t(cl['word']), cl['score'], cl['freq']) + nl)
    elif methodname == 'wipo':
        # TODO: add automatic test?
        outf.write("Extracted terms" + nl)
        outf.write("Corpus" + ': %s' % params.get('corp_full_name', '') + nl)
        outf.write(nl)
        outf.write('\t'.join(["Term", "Frequency", "Freq/mill", "Score"]) + nl)
        for k in data['Keyterms']:
            outf.write('\t'.join([t(k['item']), str(k['frq1']), str(k['relfrq1']), str(k['score'])]) + nl)
    elif methodname in ['wordlist', 'poswordlist']:
        # TODO: these normalizations shouldn't be needed in the future
        usengrams = params.get('usengrams', "0")
        usengrams = type(usengrams) == type([]) and usengrams[-1] or usengrams
        nest_ngrams = params.get('nest_ngrams', "0")
        nest_ngrams = type(nest_ngrams) == type([]) and nest_ngrams[-1] or\
                nest_ngrams
        def nested(ngram, level=0):
            if level: outf.write('"')
            for i in range(level):
                outf.write('... ')
            if level:
                outf.write('%s\t%d' % (t(ngram['str']), ngram['freq']) + nl)
            for ch in ngram.get('children', []):
                nested(ch, level+1)
        if data.get('Items', []):
            outf.write('%s\tFreq' % params.get('wlattr', '') + nl)
            for i in data['Items']:
                if params.get('wlattr', '') == 'ws_terms':
                    outf.write('%s\t%d\t' % (t(i['w2'].replace('_', ' ')),\
                            i['freq']) + nl)
                elif params.get('wlattr', '') == 'ws_collocations':
                    outf.write('%s\t%d\t%.3f' %\
                            (t(i['str']), i['freq'], float(i['score'] or '0.0'))\
                            + nl)
                else:
                    if usengrams == "1" and nest_ngrams == "1" and\
                            i.get('nested', ''):
                        continue
                    outf.write('%s\t%d' % (t(i['str']), i['freq']) + nl)
                    if usengrams == "1" and nest_ngrams == "1" and\
                            i.get('children', []):
                        nested(i)
        elif data.get('Keywords', []):
            outf.write('%s\tFreq\tFreq/mill\tFreq_ref\tFreq_ref/mill\tScore'+nl)
            for k in data['Keywords']:
                if params.get('wlattr', "0") == 'ws_terms':
                    w = t(k['str']).split('\t')[2].replace('_', ' ')
                else:
                    w = t(k['str'])
                outf.write('%s\t%d\t%.3f\t%d\t%.3f\t%.3f' %\
                        (w, k['freq'], k['rel'], k.get('freq_ref', 0),
                        k.get('rel_ref', 0.0), k.get('score', 0.0)) + nl)
    elif methodname == 'wsdiff':
        outf.write('%s\t%s' % (t(data['lbl1']), t(data['lbl2'])) + nl)
        for gr in data['content']['common']:
            h = gr['table']['Header']
            outf.write('RELATION %s\t%s\t%s\t%s\t%s' % (t(h[0]),
                    h[1], h[2], h[3], h[4]) + nl)
            for r in gr['table']['Rows']:
                outf.write('%s\t%s\t%s\t%s\t%s\t%s' % (t(r['word']),
                        r['cnt1'], r['cnt2'], r['rnk1str'], r['rnk2str'],
                        r['class']) + nl)
    elif methodname in ['extract_terms', 'extract_keywords']:
        outf.write('Item\tScore\tFreq\tRef_freq\tRel_freq\tRef_rel_freq' + nl)
        for item in data.get('terms', data.get('keywords', [])):
            outf.write('%s\t%.3f\t%d\t%d\t%.3f\t%.3f' % (t(item['item']),
                    float(item['score']), int(item['frq1']), int(item['frq2']),
                    float(item['rel_frq1']), float(item['rel_frq2'])))
            outf.write(nl)
    else:
        outf.write('Method %s does not support export to TSV.' %\
                methodname + nl)

def xml(methodname, data, outf, nl, params):
    outf.write("<?xml version='1.0' encoding='UTF-8' ?>" + nl)
    outf.write("<export>" + nl)
    outf.write("<header>" + nl)
    outf.write("  <corpus>" + n(params.get('corpname', '-')) + '</corpus>' + nl)
    outf.write("  <subcorpus>" + n(params.get('usesubcorp', '-')) +\
            '</subcorpus>' + nl)
    if data.get('Desc'):
        outf.write("  <query>" + nl)
        for d in data['Desc']:
            outf.write('    <subquery operation="%s" size="%d">%s</subquery>' %\
                    (n(d.get('op', 'operation')), d.get('size', 0),\
                    n(d.get('arg', 'argument'))) + nl)
        outf.write('  </query>' + nl)
    outf.write('</header>' + nl)
    if 'error' in data:
        outf.write('  <error>An error occurred, check results in UI or in JSON format.</error>' + nl)
        outf.write("</export>" + nl)
        return
    if methodname in ['biterms']:
        outf.write('<biterms>' + nl)
        for item in data['Biterms']:
            b = data['BitermsDict'][item][0]
            outf.write('  <item score="%.3f" co-hits="%d">' %\
                    (float(b[7]), int(b[2])) + nl)
            outf.write('    <source hits="%d">%s</source>' %\
                    (int(b[3]), n(b[0])) + nl)
            outf.write('    <target hits="%d">%s</target>' %\
                    (int(b[4]), n(b[1])) + nl)
            outf.write('  </item>' + nl)
        outf.write('</biterms>' + nl)
    elif methodname == 'wsketch':
        outf.write('<wordsketch>' + nl)
        outf.write('  <keyword pos="%s" freq="%d">%s</keyword>' %\
                (n(params.get('lpos', '')), params.get('freq', 0),
                n(params.get('lemma', ''))) + nl)
        for gr in data['Gramrels']:
            nicename = n(gr['name'].replace('"%w"', 'X').replace('%w', 'X'))
            outf.write('  <gramrel name="%s" hits="%d" score="%.3f">' %\
                    (nicename, gr['count'] or 0,
                    float(gr['score'] or '0.0') or 0.0) + nl)
            for w in gr['Words']:
                outf.write('    <coll hits="%d" score="%.3f" query="w%d">' %\
                        (w['count'], w['score'], int(w['seek'])))
                outf.write(n(w.get('word', w.get('name', ''))))
                outf.write('</coll>' + nl)
            outf.write('  </gramrel>' + nl)
        outf.write('</wordsketch>' + nl)
    elif methodname.startswith('coll'):
        mldict = {x['s']:x['n'] for x in data['Head'] if 's' in x}
        outf.write('<collocations>' + nl)
        for item in data['Items']:
            outf.write('  <item>' + nl)
            outf.write('    <str>%s</str>' % n(item['str']) + nl)
            outf.write('    <freq>%d</freq>' % item['freq'] + nl)
            outf.write('    <collfreq>%d</collfreq>' % item['coll_freq'] + nl)
            for sc in item['Stats']:
                outf.write('    <score type="%s">%.5f</score>' %\
                        (mldict.get(sc['n'], ''), float(sc['s'])) + nl)
            outf.write('  </item>' + nl)
        outf.write('</collocations>' + nl)
    elif methodname in ['view', 'first', 'concordance']:
        outf.write('<concordance>' + nl)
        def printl(ln, l):
            outf.write('  <line refs="%s" num="%d" label="%s">' %\
                    (n(','.join(l.get('Refs', []))), ln, n(l.get('linegroup', '_'))) + nl)
            left = ' '.join([n(s(x)) for x in l['Left']])
            kwic = ' '.join([n(s(x)) for x in l['Kwic']])
            right = ' '.join([n(s(x)) for x in l['Right']])
            outf.write('    <left>%s</left>' % left + nl)
            outf.write('    <kwic>%s</kwic>' % kwic + nl)
            outf.write('    <right>%s</right>' % right + nl)
            outf.write('  </line>' + nl)
        for i, line in enumerate(data['Lines']):
            if line.get('Align', []):
                outf.write('  <parallel_lines>' + nl)
                printl(i, line)
                for j, al in enumerate(line['Align']):
                    printl(j, al)
                outf.write('  </parallel_lines>' + nl)
            else:
                printl(i, line)
        outf.write('</concordance>' + nl)
    elif methodname.startswith('freq') or methodname == 'struct_wordlist':
        outf.write('<frequency>' + nl)
        for block in data['Blocks']:
            attrd = {x['s']:n(x['n']) for x in block['Head'] if 'id' in x}
            outf.write('  <items blockname="%s">' %\
                    n(block.get('blockname', '')) + nl)
            for item in block['Items']:
                outf.write('    <item>' + nl)
                for i, w in enumerate(item['Word']):
                    outf.write('      <str attribute="%s">%s</str>' %\
                            (attrd.get(i, 'str'), n(w['n']).replace('  ', ' '))
                             + nl)
                outf.write('      <freq>%d</freq>' % item['freq'] + nl)
                if 'rel' in item:
                    outf.write('      <rel>%.5f</rel>' % item['rel'] + nl)
                if 'fpm' in item:
                    outf.write('      <fpm>%.5f</fpm>' % item['fpm'] + nl)
                if params.get('examples', ''):
                    outf.write('      <example>%s</example>' %\
                            item.get('example', '') + nl)
                outf.write('    </item>' + nl)
            outf.write('  </items>' + nl)
        outf.write('</frequency>' + nl)
    elif methodname == 'thes':
        outf.write('<thesaurus>' + nl)
        outf.write('  <keyword pos="%s" freq="%d">%s</keyword>' %\
                (data.get('lpos', ''), data['freq'],
                n(data.get('lemma', ''))) + nl)
        for w in data['Words']:
            if 'Clust' in w:
                outf.write('  <cluster head="%s">' % n(w['word']) + nl)
            outf.write('  <item score="%.3f" freq="%d">%s</item>' %\
                    (w['score'], w['freq'], n(w['word'])) + nl)
            if 'Clust' in w:
                for cl in w['Clust']:
                    outf.write('  <item score="%.3f" freq="%d">%s</item>' %\
                            (cl['score'], cl['freq'], n(cl['word'])) + nl)
                outf.write('  </cluster>' + nl)
        outf.write('</thesaurus>' + nl)
    elif methodname in ['wordlist', 'poswordlist']:
        # TODO: these normalizations shouldn't be needed in the future
        usengrams = params.get('usengrams', "0")
        usengrams = type(usengrams) == type([]) and usengrams[-1] or usengrams
        nest_ngrams = params.get('nest_ngrams', "0")
        nest_ngrams = type(nest_ngrams) == type([]) and nest_ngrams[-1] or\
                nest_ngrams
        def nested(ngram, first=True):
            if not ngram.get('nested', '') or not first:
                outf.write('  <item>' + nl)
                outf.write('    <str>%s</str>' %\
                        n(ngram['str'].replace('  ', ' ')) + nl)
                outf.write('    <freq>%d</freq>' % ngram['freq'] + nl)
                for ch in ngram.get('children', []):
                    nested(ch, False)
                outf.write('  </item>' + nl)
        if data.get('Items', []):
            outf.write('<wordlist attr="%s">' %\
                    params.get('wlattr', '') + nl)
            for i in data['Items']:
                if usengrams == "1" and nest_ngrams == "1":
                    nested(i)
                outf.write('  <item>' + nl)
                if params.get('wlattr', '') == 'ws_terms':
                    outf.write('    <str>%s</str>' %\
                            n(i['w2']).replace('_', ' ') + nl)
                else:
                    outf.write('    <str>%s</str>' %\
                            n(i['str']).replace('  ', ' ') + nl)
                outf.write('    <freq>%d</freq>' % i['freq'] + nl)
                if params.get('wlattr', '') == 'ws_collocations':
                    outf.write('    <score>%s</score>' % i['score'] + nl)
                outf.write('  </item>' + nl)
            outf.write('</wordlist>' + nl)
        if data.get('Keywords', []):
            outf.write('<keywords attr="%s">' % params.get('wlattr', '') + nl)
            for k in data['Keywords']:
                if params.get('wlattr', '') == 'ws_terms':
                    w = n(k['str'].split('\t')[2]).replace('_', ' ')
                else:
                    w = n(k['str']).replace('  ', ' ')
                outf.write('  <item>' + nl)
                outf.write('    <str>%s</str>' % w + nl)
                outf.write('    <freq>%d</freq>' % k['freq'] + nl)
                outf.write('    <rel>%.3f</rel>' % k['rel'] + nl)
                outf.write('    <freq_ref>%.3f</freq_ref>' %\
                        k.get('freq_ref', 0.0) + nl)
                outf.write('    <rel_ref>%.3f</rel_ref>' %\
                        k.get('rel_ref', 0.0) + nl)
                outf.write('    <score>%.3f</score>' % k.get('score', 0.0) + nl)
                outf.write('  </item>' + nl)
            outf.write('</keywords>' + nl)
    elif methodname == 'wsdiff':
        outf.write('<wsdiff word1="%s" word2="%s">' % (n(data['lbl1']),
                n(data['lbl2'])) + nl)
        for gr in data['content']['common']:
            h = gr['table']['Header']
            outf.write('  <relation name="%s" hits1="%d" hits2="%d" '
                    'score1="%.3f" score2="%.3f">' % (n(h[0]), int(h[1]),
                    int(h[2]), float(h[3]), float(h[4])) + nl)
            for r in gr['table']['Rows']:
                outf.write('    <item hits1="%d" hits2="%d" score1="%s" '
                        'score2="%s" class="%s">%s</item>' % (int(r['cnt1']),
                        int(r['cnt2']), r['rnk1str'], r['rnk2str'],
                        r['class'], n(r['word'])) + nl)
            outf.write('  </relation>' + nl)
        outf.write('</wsdiff>' + nl)
    else:
        outf.write('<error>Method %s does not support XML export!</error>' %\
                methodname + nl)
    outf.write("</export>" + nl)

def tbx(methodname, data, outf, nl, params):
    outf.write("<?xml version='1.0' encoding='UTF-8' ?>" + nl)
    def print_prefix():
        outf.write('<martif type="TBX" xml:lang="">' + nl)
        outf.write('  <martifHeader>' + nl)
        outf.write('    <fileDesc>' + nl)
        outf.write('      <titleStmt>' + nl)
        outf.write('        <title>Sketch Engine Terminology Extraction</title>' + nl)
        outf.write('      </titleStmt>' + nl)
        outf.write('      <sourceDesc>' + nl)
        outf.write('        <p>Source corpus: %s' % data.get('corpus', '') + nl)
        outf.write('          reference corpus: %s' % data.get('ref_corpus', '') + nl)
        outf.write('      </sourceDesc>' + nl)
        outf.write('    </fileDesc>' + nl)
        outf.write('  </martifHeader>' + nl)
        outf.write('  <text>' + nl)
        outf.write('    <body>' + nl)
    def print_suffix():
        outf.write('    </body>' + nl)
        outf.write('  </text>' + nl)
        outf.write('</martif>' + nl)
    if 'error' in data:
        outf.write('      <error>An error occurred, check results in UI or in JSON format.</error>' + nl)
        print_suffix()
        return
    if methodname in ['extract_terms', 'extract_keywords']:
        print_prefix()
        for i, item in enumerate(data.get('terms', data.get('keywords', [])), 1):
            outf.write('      <termEntry id="%d">' % i + nl)
            outf.write('        <langSet xml:lang="">' + nl)
            outf.write('          <ntig>' + nl)
            outf.write('            <termGrp>' + nl)
            outf.write('              <term id="t%d">%s</term>' %\
                    (i, n(item['item'])) + nl)
            outf.write('              <termNote type="score">%.3f</termNote>' %\
                    float(item['score']) + nl)
            outf.write('              <termNote type="frequency">%d</termNote>' %\
                    int(item['frq1']) + nl)
            outf.write('              <termNote type="ref_frequency">%d</termNote>' %\
                    int(item['frq2']) + nl)
            outf.write('              <termNote type="rel_frequency">%.3f</termNote>' %\
                    float(item['rel_frq1']) + nl)
            outf.write('              <termNote type="rel_ref_frequency">%.3f</termNote>' %\
                    float(item['rel_frq2']) + nl)
            outf.write('            </termGrp>' + nl)
            outf.write('          </ntig>' + nl)
            outf.write('        </langSet>' + nl)
            outf.write('      </termEntry>' + nl)
        print_suffix()
    elif methodname == 'biterms':
        print_prefix()
        for i, item in enumerate(data['Biterms'], 1):
            b = data['BitermsDict'][item][0]
            outf.write('      <termEntry id="%d">' % i + nl)
            outf.write('        <langSet xml:lang="l1">' + nl)
            outf.write('          <ntig>' + nl)
            outf.write('            <termGrp>' + nl)
            outf.write('              <term id="t1_%d">%s</term>' %\
                    (i, n(b[0])) + nl)
            outf.write('              <termNote type="score">%.3f</termNote>' %\
                    float(b[7]) + nl)
            outf.write('              <termNote type="frequency">%d</termNote>' %\
                    int(b[2]) + nl)
            outf.write('            </termGrp>' + nl)
            outf.write('          </ntig>' + nl)
            outf.write('        </langSet>' + nl)
            outf.write('        <langSet xml:lang="l2">' + nl)
            outf.write('          <ntig>' + nl)
            outf.write('            <termGrp>' + nl)
            outf.write('              <term id="t2_%d">%s</term>' %\
                    (i, n(b[0])) + nl)
            outf.write('              <termNote type="score">%.3f</termNote>' %\
                    float(b[7]) + nl)
            outf.write('              <termNote type="frequency">%d</termNote>' %\
                    int(b[3]) + nl)
            outf.write('            </termGrp>' + nl)
            outf.write('          </ntig>' + nl)
            outf.write('        </langSet>' + nl)
            outf.write('      </termEntry>' + nl)
        print_suffix()
    else:
        outf.write('<error>Method %s does not support TBX export!</error>' %\
                methodname + nl)

def json(methodname, data, outf, nl, params):
    data['request'] = params
    jsonlib.dump(data, outf)

map = {
    'csv': csv,
    'tsv': tsv,
    'xml': xml,
    'json': json,
    'tbx': tbx,
    'xls': xls,
    'txt': txt,
    'tmx': tmx
}

def export(methodname, data, format='json', outf=sys.stdout, nl='\n', params={}):
    if format in map:
        map[format](methodname, data, outf, nl, params)
    else:
        outf.write('Export to format "%s" not supported\n' % format)
    outf.flush()
